from pytest import fixture

import numpy as np
from egttools.behaviors.CPR.abstract_cpr_strategy import AbstractCPRStrategy
from egttools.behaviors.CPR.cpr_strategies import (FairExtraction, HighExtraction, CommitmentStrategy, FakeStrategy,
                                                   FreeStrategy, NashExtraction)
from egttools.games.nonlinear_games import CommonPoolResourceDilemmaCommitment

from typing import Tuple, List


@fixture
def setup_cpr_strategy_parameters() -> Tuple[float, float, int]:
    a = 23
    b = 0.25
    N = 4

    return a, b, N


@fixture
def setup_cpr_game_parameters(setup_cpr_strategy_parameters) -> Tuple[
    int, float, float, float, float, List[AbstractCPRStrategy]]:
    a, b, N = setup_cpr_strategy_parameters
    fine = 5
    cost = 5
    strategies = [FairExtraction(), HighExtraction(), FakeStrategy(), FreeStrategy(),
                  CommitmentStrategy(1), CommitmentStrategy(2),
                  CommitmentStrategy(3), CommitmentStrategy(4)]

    return N, a, b, cost, fine, strategies


def test_nash_extraction(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters

    strategy = NashExtraction()
    assert np.round(strategy.get_extraction(a, b, N, commitment=False), 1) == 18.4
    assert np.round(strategy.get_extraction(a, b, N, commitment=True), 1) == 18.4


def test_nash_strategy_commit_proposal():
    strategy = NashExtraction()

    assert strategy.proposes_commitment() is False


def test_nash_strategy_would_like_to_commit():
    strategy = NashExtraction()

    assert strategy.would_like_to_commit() is False


def test_nash_strategy_validates_commit():
    strategy = NashExtraction()

    assert strategy.is_commitment_validated(False) is False
    assert strategy.is_commitment_validated(True) is False


def test_high_extraction(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    strategy = HighExtraction()
    assert np.round(strategy.get_extraction(a, b, N, commitment=False), 1) == 23.0
    assert np.round(strategy.get_extraction(a, b, N, commitment=True), 1) == 23.0


def test_high_strategy_commit_proposal():
    strategy = HighExtraction()

    assert strategy.proposes_commitment() is False


def test_high_strategy_would_like_to_commit():
    strategy = HighExtraction()

    assert strategy.would_like_to_commit() is False


def test_high_strategy_validates_commit():
    strategy = HighExtraction()

    assert strategy.is_commitment_validated(False) is False
    assert strategy.is_commitment_validated(True) is False


def test_fair_extraction(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    strategy = FairExtraction()
    assert np.round(strategy.get_extraction(a, b, N, commitment=False), 1) == 11.5
    assert np.round(strategy.get_extraction(a, b, N, commitment=True), 1) == 11.5


def test_fair_payoff(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    fine = 5
    cost = 5
    strategy = FairExtraction()
    assert strategy.get_payoff(a, b, strategy.get_extraction(a, b, N, True), 46, fine, cost, commitment=True) == 132.25
    assert strategy.get_payoff(a, b, strategy.get_extraction(a, b, N, True), 46, fine, cost, commitment=False) == 132.25


def test_fair_strategy_commit_proposal():
    strategy = FairExtraction()

    assert strategy.proposes_commitment() is False


def test_fair_strategy_would_like_to_commit():
    strategy = FairExtraction()

    assert strategy.would_like_to_commit() is True


def test_fair_strategy_validates_commit():
    strategy = FairExtraction()

    assert strategy.is_commitment_validated(False) is False
    assert strategy.is_commitment_validated(True) is True


def test_commitment_strategy_threshold():
    strategy = CommitmentStrategy(3)

    assert strategy.is_commitment_validated(0) is False
    assert strategy.is_commitment_validated(2) is False
    assert strategy.is_commitment_validated(3) is True
    assert strategy.is_commitment_validated(4) is True


def test_commitment_strategy_extraction(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    strategy = CommitmentStrategy(3)

    assert strategy.get_extraction(a, b, N, commitment=False) == 23.0
    assert strategy.get_extraction(a, b, N, commitment=True) == 11.5


def test_commitment_strategy_commit_proposal():
    strategy = CommitmentStrategy(3)

    assert strategy.proposes_commitment() is True


def test_commitment_strategy_would_like_to_commit():
    strategy = CommitmentStrategy(3)

    assert strategy.would_like_to_commit() is True


def test_commitment_strategy_validates_commit():
    strategy = CommitmentStrategy(3)

    assert strategy.is_commitment_validated(0) is False
    assert strategy.is_commitment_validated(1) is False
    assert strategy.is_commitment_validated(2) is False
    assert strategy.is_commitment_validated(3) is True
    assert strategy.is_commitment_validated(4) is True


def test_fake_strategy(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    strategy = FakeStrategy()

    assert strategy.get_extraction(a, b, N, commitment=False) == 23.0
    assert strategy.get_extraction(a, b, N, commitment=True) == 23.0


def test_fake_strategy_commit_proposal():
    strategy = FakeStrategy()

    assert strategy.proposes_commitment() is False


def test_fake_strategy_would_like_to_commit():
    strategy = FakeStrategy()

    assert strategy.would_like_to_commit() is True


def test_fake_strategy_validates_commit():
    strategy = FakeStrategy()

    assert strategy.is_commitment_validated(False) is False
    assert strategy.is_commitment_validated(True) is True


def test_free_strategy(setup_cpr_strategy_parameters):
    a, b, N = setup_cpr_strategy_parameters
    strategy = FreeStrategy()

    assert strategy.get_extraction(a, b, N, commitment=False) == 23.0
    assert strategy.get_extraction(a, b, N, commitment=True) == 11.5


def test_free_strategy_commit_proposal():
    strategy = FreeStrategy()

    assert strategy.proposes_commitment() is False


def test_free_strategy_would_like_to_commit():
    strategy = FreeStrategy()

    assert strategy.would_like_to_commit() is True


def test_free_strategy_validates_commit():
    strategy = FreeStrategy()

    assert strategy.is_commitment_validated(False) is False
    assert strategy.is_commitment_validated(True) is True


def test_cpr_game_play(setup_cpr_game_parameters):
    # There are 8 strategies:
    # Fair, High, Fake, Free, COM1, COM2, COM3, COM4
    group_size, a, b, cost, fine, strategies = setup_cpr_game_parameters
    game = CommonPoolResourceDilemmaCommitment(group_size, a, b, cost, fine, strategies)

    game_payoffs = np.zeros(shape=(game.nb_strategies(),))
    # Case of all 3 Fair 1 High
    group_composition = [3, group_size - 3, 0, 0, 0, 0, 0, 0]

    game.play(group_composition, game_payoffs)

    assert game_payoffs[0] == 99.1875
    assert game_payoffs[1] == 198.375

    game_payoffs.fill(0)

    # Case of all 3 Fair 1 High
    group_composition = [2, 1, 1, 0, 0, 0, 0, 0]

    game.play(group_composition, game_payoffs)

    assert game_payoffs[0] == 66.125
    assert game_payoffs[1] == 132.25
    assert game_payoffs[2] == 132.25


def test_cpr_calculate_payoffs(setup_cpr_game_parameters):
    pass


def test_cpr_calculate_total_extraction(setup_cpr_game_parameters):
    # There are 8 strategies:
    # Fair, High, Fake, Free, COM1, COM2, COM3, COM4
    group_size, a, b, cost, fine, strategies = setup_cpr_game_parameters
    game = CommonPoolResourceDilemmaCommitment(group_size, a, b, cost, fine, strategies)

    commitment_accepted = np.zeros(shape=(game.nb_strategies(),), dtype=bool)

    # Case of all Fair
    group_composition = [group_size, 0, 0, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 46

    commitment_accepted.fill(0)

    # Case of all High
    group_composition = [0, group_size, 0, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 92

    commitment_accepted.fill(0)

    # Case of all Fake
    group_composition = [0, 0, group_size, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 92

    commitment_accepted.fill(0)

    # Case of all Free
    group_composition = [0, 0, 0, group_size, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 92

    commitment_accepted.fill(0)

    # Case of all COM1
    group_composition = [0, 0, 0, 0, group_size, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 46

    commitment_accepted.fill(0)

    # Case of all COM2
    group_composition = [0, 0, 0, 0, 0, group_size, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 46

    commitment_accepted.fill(0)

    # Case of all COM3
    group_composition = [0, 0, 0, 0, 0, 0, group_size, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 46

    commitment_accepted.fill(0)

    # Case of all COM4
    group_composition = [0, 0, 0, 0, 0, 0, 0, group_size]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 46

    commitment_accepted.fill(0)

    # Case of all 3 Fair 1 High
    group_composition = [3, group_size - 3, 0, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert all(commitment_accepted == 0)
    assert game.calculate_total_extraction(commitment_accepted, group_composition) == 57.5


def test_cpr_get_nb_committed(setup_cpr_game_parameters):
    # There are 8 strategies:
    # Fair, High, Fake, Free, COM1, COM2, COM3, COM4

    group_size, a, b, cost, fine, strategies = setup_cpr_game_parameters
    game = CommonPoolResourceDilemmaCommitment(group_size, a, b, cost, fine, strategies)

    # Case of all Fair
    group_composition = [group_size, 0, 0, 0, 0, 0, 0, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all High
    group_composition = [0, group_size, 0, 0, 0, 0, 0, 0]
    assert game.get_nb_committed(group_composition) == 0

    # Case of all Fake
    group_composition = [0, 0, group_size, 0, 0, 0, 0, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all Free
    group_composition = [0, 0, 0, group_size, 0, 0, 0, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all COM1
    group_composition = [0, 0, 0, 0, group_size, 0, 0, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all COM2
    group_composition = [0, 0, 0, 0, 0, group_size, 0, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all COM2
    group_composition = [0, 0, 0, 0, 0, 0, group_size, 0]
    assert game.get_nb_committed(group_composition) == group_size

    # Case of all COM4
    group_composition = [0, 0, 0, 0, 0, 0, 0, group_size]
    assert game.get_nb_committed(group_composition) == group_size

    # Let's also check a mix of High and another strategy
    group_composition = [0, group_size - 1, 0, 0, 0, 0, 0, 1]
    assert game.get_nb_committed(group_composition) == 1


def test_cpr_check_if_commitment_validated(setup_cpr_game_parameters):
    # There are 8 strategies:
    # Fair, High, Fake, Free, COM1, COM2, COM3, COM4

    group_size, a, b, cost, fine, strategies = setup_cpr_game_parameters
    game = CommonPoolResourceDilemmaCommitment(group_size, a, b, cost, fine, strategies)

    commitment_accepted = np.zeros(shape=(game.nb_strategies(),), dtype=bool)

    # Case of all Fair
    group_composition = [group_size, 0, 0, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert np.all(commitment_accepted == 0)

    commitment_accepted.fill(0)

    # Case of all High
    group_composition = [0, group_size, 0, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert np.all(commitment_accepted == 0)

    commitment_accepted.fill(0)

    # Case of all Fake
    group_composition = [0, 0, group_size, 0, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert np.all(commitment_accepted == 0)

    commitment_accepted.fill(0)

    # Case of all Free
    group_composition = [0, 0, 0, group_size, 0, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert np.all(commitment_accepted == 0)

    commitment_accepted.fill(0)

    # Case of all COM1
    group_composition = [0, 0, 0, 0, group_size, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert commitment_accepted[4]

    commitment_accepted.fill(0)

    # Case of all COM2
    group_composition = [0, 0, 0, 0, 0, group_size, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert commitment_accepted[5]

    commitment_accepted.fill(0)

    # Case of all COM3
    group_composition = [0, 0, 0, 0, 0, 0, group_size, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert commitment_accepted[6]

    commitment_accepted.fill(0)

    # Case of all COM4
    group_composition = [0, 0, 0, 0, 0, 0, 0, group_size]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert commitment_accepted[7]

    commitment_accepted.fill(0)

    # Let's also check a mix of High and another strategy
    group_composition = [0, group_size - 1, 0, 0, 0, 0, 0, 1]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[7]
    assert not commitment_accepted[1]

    commitment_accepted.fill(0)

    # Let's check if 1 COM1 is present with the rest High
    group_composition = [0, group_size - 1, 0, 0, 1, 0, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[1]
    assert commitment_accepted[4]

    commitment_accepted.fill(0)

    # Let's check if 1 COM2 is present with the rest High
    group_composition = [0, group_size - 1, 0, 0, 0, 1, 0, 0]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[1]
    assert not commitment_accepted[5]

    commitment_accepted.fill(0)

    # Let's check if 1 COM3 is present with the rest High
    group_composition = [0, group_size - 1, 0, 0, 0, 0, 0, 1]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[1]
    assert not commitment_accepted[6]

    commitment_accepted.fill(0)

    # Let's check if 1 COM4 is present with the rest High
    group_composition = [0, group_size - 1, 0, 0, 0, 0, 0, 1]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[1]
    assert not commitment_accepted[7]

    commitment_accepted.fill(0)

    # Let's check if 1 COM4 with Fair, Fake and Free
    group_composition = [1, 0, 1, 1, 0, 0, 0, group_size - 3]
    nb_committed = game.get_nb_committed(group_composition)
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert commitment_accepted[0]
    assert commitment_accepted[2]
    assert commitment_accepted[3]
    assert commitment_accepted[7]

    commitment_accepted.fill(0)

    # Let's check if 1 COM4 with High, Fake and Free
    group_composition = [0, 1, 1, 1, 0, 0, 0, group_size - 3]
    nb_committed = game.get_nb_committed(group_composition)
    assert nb_committed == 3
    game.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
    assert not commitment_accepted[1]
    assert not commitment_accepted[2]
    assert not commitment_accepted[3]
    assert not commitment_accepted[7]

# def test_fair_payoff_from_game(setup_cpr_game_parameters):
#     # There are 8 strategies:
#     # Fair, High, Fake, Free, COM1, COM2, COM3, COM4
#
#     group_size, a, b, cost, fine, strategies = setup_cpr_game_parameters
#     game = CommonPoolResourceDilemmaCommitment(group_size, a, b, cost, fine, strategies)
#
#     commitment_accepted = np.zeros(shape=(game.nb_strategies(),), dtype=bool)
#
#     strategy = FairExtraction()
#
#     game.grou
#
#     assert strategy.get_payoff(a, b, strategy.get_extraction(a, b, group_size, True), 46, fine, cost, commitment=True) == 132.25
#     assert strategy.get_payoff(a, b, strategy.get_extraction(a, b, group_size, True), 46, fine, cost, commitment=False) == 132.25



"""
API reference documentation for the `analytical` submodule.
"""

try:
    from ..numerical.numerical_ import PairwiseComparison, replicator_equation_n_player
except Exception:
    raise Exception("numerical package not initialized")
else:
    from .sed_analytical import replicator_equation
    from .sed_analytical import StochDynamics

__all__ = ['replicator_equation', 'StochDynamics', 'PairwiseComparison', 'replicator_equation_n_player']



"""
API reference documentation for the `games` submodule.
"""

try:
    from ..numerical.numerical_.games import (AbstractGame,
                                              AbstractSpatialGame,
                                              AbstractNPlayerGame,
                                              NormalFormGame,
                                              NormalFormNetworkGame,
                                              CRDGame,
                                              CRDGameTU,
                                              OneShotCRD,
                                              OneShotCRDNetworkGame,
                                              Matrix2PlayerGameHolder,
                                              MatrixNPlayerGameHolder, )
except Exception:
    raise Exception("numerical package not initialized")
else:
    from .pgg import PGG
    from .informal_risk import InformalRiskGame
    from .abstract_games import AbstractTwoPLayerGame
    from .nonlinear_games import NPlayerStagHunt, CommonPoolResourceDilemma, CommonPoolResourceDilemmaCommitment

__all__ = ['AbstractGame', 'AbstractSpatialGame', 'AbstractNPlayerGame', 'NormalFormGame', 'NormalFormNetworkGame',
           'CRDGame', 'CRDGameTU',
           'OneShotCRD', 'OneShotCRDNetworkGame', 'PGG',
           'InformalRiskGame', 'AbstractTwoPLayerGame', 'NPlayerStagHunt', 'CommonPoolResourceDilemma',
           'CommonPoolResourceDilemmaCommitment', 'Matrix2PlayerGameHolder', 'MatrixNPlayerGameHolder']



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""
This python module contains the necessary functions
to calculate analytically the evolutionary dynamics in Infinite and Finite
populations on 2-player games.
"""

import numpy as np
import numpy.typing as npt
from scipy.sparse import lil_matrix, csr_matrix
from scipy.linalg import eig
from scipy.stats import hypergeom, multivariate_hypergeom, multinomial
from itertools import permutations
from typing import Tuple, Optional
from warnings import warn
from .. import sample_simplex, calculate_nb_states, calculate_state


def replicator_equation(x: np.ndarray, payoffs: np.ndarray) -> np.ndarray:
    """
    Produces the discrete time derivative of the replicator dynamics

    This only works for 2-player games.

    Parameters
    ----------
    x : numpy.ndarray[numpy.float64[m,1]]
        array containing the frequency of each strategy in the population.
    payoffs : numpy.ndarray[numpy.float64[m,m]]
        payoff matrix
    Returns
    -------
    numpy.ndarray
        time derivative of x

    See Also
    --------
    egttools.analytical.StochDynamics
    egttools.numerical.PairwiseComparisonNumerical
    """
    ax = np.dot(payoffs, x)
    return x * (ax - np.dot(x, ax))


def replicator_equation_n_player(x: np.ndarray, payoffs: np.ndarray, group_size: int) -> np.ndarray:
    """
    Replicator dynamics in N-player games

    The replicator equation is of the form

    .. math::
        g(x) \\equiv \\dot{x_{i}} = x_{i}(f_{i}(x) - \\sum_{j=1}^{N}{x_{j}f_{j}(x))

    Which can also be represented using a pairwise comparison rule as:

    .. math::
        \\dot{x_{i}} = x_{i}\\sum_{j}(f_{ij}(x) - f_{ji}(x))x_{j}

    For N-player games, to calculate the fitness of a strategy given a population state, we
    need to calculate the probability of each possible group configuration. This can be obtained
    by summing for each possible group configuration the payoff of strategy i times the probability
    of the group configurations occurring.

    Parameters
    ----------
    x : numpy.ndarray
        A vector of shape (1, nb_strategies), which contains the current frequency of each strategy in the population.
    payoffs : numpy.ndarray
        Payoff matrix. Each row represents a strategy and each column a possible group configuration.
        Each entry in the matrix should give the expected payoff for each row strategy for a given column group
        configuration.
    group_size : int
        Size of the group.

    Returns
    -------
    numpy.ndarray
        A vector of shape (1, nb_strategies), which contains the change in frequency of each strategy in the population
        (so the gradient).

    """
    fitness = np.zeros(shape=(len(x),))
    fitness_avg = 0.
    nb_group_configurations = calculate_nb_states(group_size, len(x))
    for strategy_index in range(len(x)):
        for i in range(nb_group_configurations):
            group_configuration = sample_simplex(i, group_size, len(x))
            if group_configuration[strategy_index] > 0:
                group_configuration[strategy_index] -= 1
                prob = multinomial.pmf(group_configuration, group_size - 1, x)
                fitness[strategy_index] += prob * payoffs[strategy_index, i]
        fitness_avg += x[strategy_index] * fitness[strategy_index]

    return x * (fitness - fitness_avg)


class StochDynamics:
    """A class containing methods to calculate the stochastic evolutionary dynamics of a population.

    Defines a class that contains methods to compute the stationary distribution for
    the limit of small mutation (only the monomorphic states) and the full transition matrix.

    Parameters
    ----------
    nb_strategies : int
                number of strategies in the population
    payoffs : numpy.ndarray[numpy.float64[m,m]]
            Payoff matrix indicating the payoff of each strategy (rows) against each other (columns).
            When analyzing an N-player game (group_size > 2) the structure of the matrix is a bit more involved,
            and we can have 2 options for structuring the payoff matrix:

            1) If we consider a simplified version of the system with a reduced Markov Chain which only contains
            the states at the edges of the simplex (the Small Mutation Limit - SML), then, we can assume that, at most,
            there will be 2 strategies in a group at any given moment. In this case, StochDynamics expects
            a square matrix of size nb_strategies x nb_strategies, in which each entry is a function that takes
            2 positional arguments k and group_size, and an optional *args argument, and will return the expected payoff
            of the row strategy A in a group with k A strategists and group_size - k
            B strategists (the column strategy). For all the elements in the diagonal, only 1 strategy should be present
            in the group, thus, this function should always return the same value, i.e., the payoff of a row strategy
            when all individuals in the group adopt the same strategy. See below for an example.

            2) If we want to consider the full Markov Chain composed of all possible states in the simplex, then
            the payoff matrix should be of the shape nb_strategies x nb_group_configurations, where the number
            of group configurations can be calculated using `egttools.calculate_nb_states(group_size, nb_strategies)`.
            Moreover, the mapping between group configurations and integer indexes must be done using
            `egttools.sample_simplex(index, group_size, nb_strategies)`. See below for an example
    pop_size : int
            population size
    group_size : int
            group size
    mu : float
        mutation probability

    See Also
    --------
    egttools.numerical.PairwiseComparisonNumerical
    egttools.analytical.replicator_equation
    egttools.analytical.PairwiseComparison

    Notes
    -----
    We recommend that instead of`StochDynamics`, you use `PairwiseComparison` because the latter
    is implemented in C++, runs faster and supports more precise types.

    Examples
    --------
    Example of the payoff matrix for case 1) mu = 0:
        >>> def get_payoff_a_vs_b(k, group_size, *args):
        ...     pre_computed_payoffs = [4, 5, 2, ..., 4] # the size of this list should be group_size + 1
        ...     return pre_computed_payoffs[k]
        >>> def get_payoff_b_vs_a(k, group_size, *args):
        ...     pre_computed_payoffs = [0, 2, 1, ..., 0] # the size of this list should be group_size + 1
        ...     return pre_computed_payoffs[k]
        >>> def get_payoff_a_vs_a(k, group_size, *args):
        ...     pre_computed_payoffs = [1, 1, 1, ..., 1] # the size of this list should be group_size + 1
        ...     return pre_computed_payoffs[k]
        >>> def get_payoff_b_vs_b(k, group_size, *args):
        ...     pre_computed_payoffs = [0, 0, 0, ..., 0] # the size of this list should be group_size + 1
        ...     return pre_computed_payoffs[k]
        >>> payoff_matrix = np.array([
        ...     [get_payoff_A_vs_A, get_payoff_A_vs_B],
        ...     [get_payoff_B_vs_A, get_payoff_B_vs_B]
        ...     ])

    Example of payoff matrix for case 2) full markov chain (mu > 0):
        >>> import egttools
        >>> nb_group_combinations = egttools.calculate_nb_states(group_size, nb_strategies)
        >>> payoff_matrix = np.zeros(shape=(nb_strategies, nb_group_combinations))
        >>> for group_configuration_index in range(nb_group_combinations):
        ...     for strategy in range(nb_strategies):
        ...         group_configuration = egttools.sample_simplex(group_configuration_index, group_size, nb_strategies)
        ...         payoff_matrix[strategy, group_configuration_index] = get_payoff(strategy, group_configuration)
    """

    def __init__(self, nb_strategies: int, payoffs: np.ndarray, pop_size: int, group_size=2, mu=0) -> None:
        self.nb_strategies = nb_strategies
        self.payoffs = payoffs
        self.pop_size = pop_size
        self.group_size = group_size
        self.mu = mu
        self.nb_states_population = calculate_nb_states(pop_size, nb_strategies)
        self.nb_group_combinations = calculate_nb_states(group_size, nb_strategies)
        if group_size > 2:  # pairwise game
            self.fitness = self.fitness_group
            self.full_fitness = self.full_fitness_difference_group
        else:  # group game
            self.fitness = self.fitness_pair
            self.full_fitness = self.full_fitness_difference_pairwise

    def update_population_size(self, pop_size: int):
        """
        Updates the size of the population and the number of possible population states.

        Parameters
        ----------
        pop_size: New population size
        """
        self.pop_size = pop_size
        self.nb_states_population = calculate_nb_states(pop_size, self.nb_strategies)

    def update_group_size(self, group_size: int):
        """
        Updates the groups size of the game (and the methods used to compute the fitness)

        Parameters
        ----------
        group_size: new group size
        """
        self.group_size = group_size
        self.nb_group_combinations = calculate_nb_states(group_size, self.nb_strategies)
        if group_size > 2:  # pairwise game
            self.fitness = self.fitness_group
            self.full_fitness = self.full_fitness_difference_group
        else:  # group game
            self.fitness = self.fitness_pair
            self.full_fitness = self.full_fitness_difference_pairwise

    def update_payoffs(self, payoffs: np.ndarray, nb_strategies: Optional[int] = None):
        """
        Updates the payoff matrix

        Parameters
        ----------
        payoffs: payoff matrix
        nb_strategies: total number of strategies (optional). If not indicated, then the new payoff
                       matrix must have the same dimensions as the previous one
        """
        if nb_strategies is None:
            if payoffs.shape[0] != self.nb_strategies:
                raise ValueError("The number of rows of the payoff matrix must be equal to the number of strategies.")
        else:
            self.nb_strategies = nb_strategies
        self.payoffs = payoffs

    def fitness_pair(self, x: int, i: int, j: int, *args: Optional[list]) -> float:
        """
        Calculates the fitness of strategy i versus strategy j, in
        a population of x i-strategists and (pop_size-x) j strategists, considering
        a 2-player game.

        Parameters
        ----------
        x : int
            number of i-strategists in the population
        i : int
            index of strategy i
        j : int
            index of strategy j
        args : Optional[list]

        Returns
        -------
            float
            the fitness difference among the strategies
        """
        fitness_i = ((x - 1) * self.payoffs[i, i] +
                     (self.pop_size - x) * self.payoffs[i, j]) / (self.pop_size - 1)
        fitness_j = ((self.pop_size - x - 1) * self.payoffs[j, j] +
                     x * self.payoffs[j, i]) / (self.pop_size - 1)
        return fitness_i - fitness_j

    def full_fitness_difference_pairwise(self, i: int, j: int, population_state: np.ndarray) -> float:
        """
        Calculates the fitness of strategy i in a population with state :param population_state,
        assuming pairwise interactions (2-player game).

        Parameters
        ----------
        i : int
            index of the strategy that will reproduce
        j : int
            index of the strategy that will die
        population_state : numpy.ndarray[numpy.int64[m,1]]
                           vector containing the counts of each strategy in the population

        Returns
        -------
        float
        The fitness difference between the two strategies for the given population state
        """
        fitness_i = (population_state[i] - 1) * self.payoffs[i, i]
        for strategy in range(self.nb_strategies):
            if strategy == i:
                continue
            fitness_i += population_state[strategy] * self.payoffs[i, strategy]
        fitness_j = (population_state[j] - 1) * self.payoffs[j, j]
        for strategy in range(self.nb_strategies):
            if strategy == j:
                continue
            fitness_j += population_state[strategy] * self.payoffs[j, strategy]

        return (fitness_i - fitness_j) / (self.pop_size - 1)

    def fitness_group(self, x: int, i: int, j: int, *args: Optional[list]) -> float:
        """
        In a population of x i-strategists and (pop_size-x) j strategists, where players
        interact in group of 'group_size' participants this function
        returns the average payoff of strategies i and j. This function expects
        that

        .. math::
            x \\in [1,pop_size-1]

        Parameters
        ----------
        x : int
            number of individuals adopting strategy i in the population
        i : int
            index of strategy i
        j : int
            index of strategy j
        args : Optional[list]
            Other Parameters. This can be used to pass extra parameters to functions
            stored in the payoff matrix

        Returns
        -------
            float
            Returns the difference in fitness between strategy i and j
        """
        k_array_1 = np.arange(0, self.group_size, dtype=np.int64)
        k_array_2 = np.arange(0, self.group_size, dtype=np.int64)
        i_pmf = hypergeom(self.pop_size - 1, x - 1, self.group_size - 1).pmf(k_array_1)
        j_pmf = hypergeom(self.pop_size - 1, x, self.group_size - 1).pmf(k_array_2)

        fitness_i, fitness_j = 0, 0
        for k in k_array_1:
            fitness_i += self.payoffs[i, j](k + 1, self.group_size, *args) * i_pmf[k]
            fitness_j += self.payoffs[j, i](self.group_size - k, self.group_size, *args) * j_pmf[k]

        return fitness_i - fitness_j

    def full_fitness_difference_group(self, i: int, j: int, population_state: np.ndarray) -> float:
        """
        Calculate the fitness difference between strategies :param i and :param j
        assuming that player interacts in groups of size group_size > 2 (n-player games).

        Parameters
        ----------
        i : int
            index of the strategy that will reproduce
        j : int
            index of the strategy that will die
        population_state : numpy.ndarray[numpy.int64[m,1]]
                           vector containing the counts of each strategy in the population

        Returns
        -------
        float
        The fitness difference between strategies i and j
        """
        copy1 = population_state.copy()
        copy1[i] -= 1
        copy2 = population_state.copy()
        copy2[j] -= 1
        rv_i = multivariate_hypergeom(copy1, self.group_size - 1)
        rv_j = multivariate_hypergeom(copy2, self.group_size - 1)

        fitness_i, fitness_j = 0., 0.
        for group_index in range(self.nb_group_combinations):
            group = sample_simplex(group_index, self.group_size, self.nb_strategies)
            if group[i] > 0:
                group[i] -= 1
                fitness_i += self.payoffs[i, group_index] * rv_i.pmf(group)
                group[i] += 1
            if group[j] > 0:
                group[j] -= 1
                fitness_j += self.payoffs[j, group_index] * rv_j.pmf(group)
                group[j] += 1

        return fitness_i - fitness_j

    @staticmethod
    def fermi(beta: float, fitness_diff: float) -> npt.ArrayLike:
        """
        The fermi function determines the probability that the first type imitates the second.

        Parameters
        ----------
        beta : float
            intensity of selection
        fitness_diff : float
            Difference in fitness between the strategies (f_a - f_b).

        Returns
        -------
        numpy.typing.ArrayLike
            the probability of imitation
        """
        return np.clip(1. / (1. + np.exp(beta * fitness_diff, dtype=np.float64)), 0., 1.)

    def prob_increase_decrease(self, k: int, invader: int, resident: int,
                               beta: float, *args: Optional[list]) -> Tuple[npt.ArrayLike, npt.ArrayLike]:
        """
        This function calculates for a given number of invaders the probability
        that the number increases or decreases with one.

        Parameters
        ----------
        k : int
            number of invaders in the population
        invader: int
            index of the invading strategy
        resident: int
            index of the resident strategy
        beta: float
            intensity of selection
        args: Optional[list]
            other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.
        Returns
        -------
        Tuple[numpy.typing.ArrayLike, numpy.typing.ArrayLike]
            tuple(probability of increasing the number of invaders, probability of decreasing)
        """
        if (k == self.pop_size) or (k == 0):
            increase = 0
            decrease = 0
        else:
            fitness_diff = self.fitness(k, invader, resident, *args)
            increase = (((self.pop_size - k) / self.pop_size) *
                        (k / (self.pop_size - 1))) * StochDynamics.fermi(-beta,
                                                                         fitness_diff)

            decrease = ((k / self.pop_size) * ((self.pop_size - k) /
                                               (self.pop_size - 1))) * StochDynamics.fermi(beta,
                                                                                           fitness_diff)
        return np.clip(increase, 0., 1.), np.clip(decrease, 0., 1.)

    def prob_increase_decrease_with_mutation(self, k: int, invader: int, resident: int, beta: float,
                                             *args: Optional[list]) -> Tuple[float, float]:
        """
        This function calculates for a given number of invaders the probability
        that the number increases or decreases with taking into account a mutation rate.

        Parameters
        ----------
        k : int
            number of invaders in the population
        invader: int
            index of the invading strategy
        resident: int
            index of the resident strategy
        beta: float
            intensity of selection
        args: Optional[list]
            other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.
        Returns
        -------
        Tuple[float, float]
            tuple(probability of increasing the number of invaders, probability of decreasing)
        """
        p_plus, p_less = self.prob_increase_decrease(k, invader, resident, beta, *args)
        p_plus = ((1 - self.mu) * p_plus) + (self.mu * ((self.pop_size - k) / self.pop_size))
        p_less = ((1 - self.mu) * p_less) + (self.mu * (k / self.pop_size))
        return p_plus, p_less

    def gradient_selection(self, k: int, invader: int, resident: int, beta: float, *args: Optional[list]) -> float:
        """
        Calculates the gradient of selection given an invader and a resident strategy.

        Parameters
        ----------
        k : int
            number of invaders in the population
        invader : int
            index of the invading strategy
        resident : int
            index of the resident strategy
        beta : float
            intensity of selection
        args : Optional[List]
            other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.

        Returns
        -------
        float
            The gradient of selection.
        """
        if k == 0:
            return 0
        elif k == self.pop_size:
            return 0
        else:
            return ((self.pop_size - k) / self.pop_size) * (k / (self.pop_size - 1)) * np.tanh(
                (beta / 2) * self.fitness(k, invader, resident, *args))

    def full_gradient_selection(self, population_state: np.ndarray, beta: float) -> np.ndarray:
        """
        Calculates the gradient of selection for an invading strategy, given a population state.

        Parameters
        ----------
        population_state : numpy.ndarray[np.int64[m,1]]
            structure of unsigned integers containing the
            counts of each strategy in the population
        beta : float
            intensity of selection

        Returns
        -------
        numpy.ndarray[numpy.float64[m,m]]
            Matrix indicating the likelihood of change in the population given a starting point.
        """
        probability_selecting_strategy_first = population_state / self.pop_size
        probability_selecting_strategy_second = population_state / self.pop_size
        probabilities = np.outer(probability_selecting_strategy_first, probability_selecting_strategy_second)
        fitness = np.zeros(shape=(self.nb_strategies, self.nb_strategies))
        for j in range(self.nb_strategies):
            if population_state[j] == 0:
                continue
            for i in range(self.nb_strategies):
                if population_state[i] == 0:
                    continue
                fitness[j, i] = self.full_fitness(i, j, population_state)

        return (probabilities * np.tanh((beta / 2) * fitness)).sum(axis=0) * (1 - self.mu) + (
                self.mu / (self.nb_strategies - 1)) * probability_selecting_strategy_second

    def full_gradient_selection_without_mutation(self, population_state: np.ndarray, beta: float) -> np.ndarray:
        """
        Calculates the gradient of selection for an invading strategy, given a population state. It does
        not take into account mutation.

        Parameters
        ----------
        population_state : numpy.ndarray[np.int64[m,1]]
            structure of unsigned integers containing the
            counts of each strategy in the population
        beta : float
            intensity of selection

        Returns
        -------
        numpy.ndarray[numpy.float64[m,m]]
            Matrix indicating the likelihood of change in the population given a starting point.
        """

        probability_selecting_strategy_first = population_state / self.pop_size
        probability_selecting_strategy_second = population_state / self.pop_size
        probabilities = np.outer(probability_selecting_strategy_first, probability_selecting_strategy_second)
        fitness = np.zeros(shape=(self.nb_strategies, self.nb_strategies))
        for j in range(self.nb_strategies):
            if population_state[j] == 0:
                continue
            for i in range(self.nb_strategies):
                if population_state[i] == 0:
                    continue
                fitness[j, i] = self.full_fitness(i, j, population_state)

        return (probabilities * np.tanh((beta / 2) * fitness)).sum(axis=0)

    def fixation_probability(self, invader: int, resident: int, beta: float, *args: Optional[list]) -> float:
        """
        Function for calculating the fixation_probability probability of the invader
        in a population of residents.

        TODO: Requires more testing!

        Parameters
        ----------
        invader : int
            index of the invading strategy
        resident : int
            index of the resident strategy
        beta : float
            intensity of selection
        args : Optional[list]
            Other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.

        Returns
        -------
        float
            The fixation_probability probability.

        See Also
        --------
        egttools.numerical.PairwiseComparisonNumerical
        """
        phi = 0.
        prod = 1.
        for i in range(1, self.pop_size):
            p_plus, p_minus = self.prob_increase_decrease(i, invader, resident, beta, *args)
            # this is necessary to avoid divisions by zero
            if np.isclose(p_plus, 0., atol=1e-12) and not np.isclose(p_plus, p_minus):
                return 0.
            prod *= p_minus / p_plus
            phi += prod
            # We can approximate by zero if phi is too big
            if phi > 1e7:
                return 0.0

        return 1.0 / (1.0 + phi)

    def calculate_full_transition_matrix(self, beta: float, *args: Optional[list]) -> csr_matrix:
        """
        Returns the full transition matrix in sparse representation.

        Parameters
        ----------
        beta : float
            Intensity of selection.
        args : Optional[list]
            Other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.

        Returns
        -------
        scipy.sparse.csr_matrix
            The full transition matrix between the two strategies in sparse format.
        """
        nb_states = calculate_nb_states(self.pop_size, self.nb_strategies)
        mutation_probability = (self.mu / (self.nb_strategies - 1))
        not_mu = 1. - self.mu

        transitions = lil_matrix((nb_states, nb_states), dtype=np.float64)
        possible_transitions = [1, -1]
        for i in range(self.nb_strategies - 2):
            possible_transitions.append(0)

        for i in range(nb_states):
            total_prob = 0.
            current_state = sample_simplex(i, self.pop_size, self.nb_strategies)
            # Check if we are in a monomorphic state
            monomorphic = True if (current_state == self.pop_size).any() else False

            # calculate probability of transitioning from current_tate to next_state
            for permutation in permutations(possible_transitions):
                # get new state
                new_state = current_state + permutation

                # Check if we are trying an impossible transition
                if (new_state < 0).any() or (new_state > self.pop_size).any():
                    continue

                new_state_index = calculate_state(self.pop_size, new_state)

                # If we are in a monomorphic population, transitions
                # can only happen if a mutation event occurs
                if monomorphic:
                    transitions[i, new_state_index] = mutation_probability
                else:
                    increase = np.where(np.array(permutation) == 1)[0][0]
                    decrease = np.where(np.array(permutation) == -1)[0][0]

                    if current_state[increase] == 0:
                        prob = (current_state[decrease] / self.pop_size) * mutation_probability
                    else:
                        # now we calculate the transition probability
                        fitness_diff = self.full_fitness(decrease, increase, current_state)
                        # Probability that the individual that will die is selected and that the individual that
                        # will be imitated is selected times the probability of imitation
                        prob = not_mu * (current_state[increase] / (self.pop_size - 1))
                        prob *= StochDynamics.fermi(beta, fitness_diff)
                        # The probability that there will not be a mutation event times the probability
                        # of the transition
                        # plus the probability that if there is a mutation event, the dying strategy is selected
                        # times the probability that it mutates into the increasing strategy
                        prob = (current_state[decrease] / self.pop_size) * (prob + mutation_probability)
                        total_prob += prob

                    transitions[i, new_state_index] = prob

            if monomorphic:
                transitions[i, i] = not_mu
            else:
                transitions[i, i] = 1. - total_prob

        return transitions.tocsr().transpose()

    def transition_and_fixation_matrix(self, beta: float, *args: Optional[list]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the transition matrix (only for the monomorphic states)
        and the fixation_probability probabilities.

        This method calculates the transitions between monomorphic states. Thus, it assumes
        that we are in the small mutation limit (SML) of the moran process. Only
        use this method if this assumption is reasonable.

        Parameters
        ----------
        beta : float
            Intensity of selection.
        args : Optional[list]
            Other arguments. Can be used to pass extra arguments to functions contained
            in the payoff matrix.
        Returns
        -------
        Tuple[numpy.ndarray[numpy.float64[m,m]], numpy.ndarray[numpy.float64[m,m]]]
            This method returns a tuple with the transition matrix as first element, and
            the matrix of fixation probabilities.
        """
        transitions = np.zeros((self.nb_strategies, self.nb_strategies))
        fixation_probabilities = np.zeros((self.nb_strategies, self.nb_strategies))

        for first in range(self.nb_strategies):
            transitions[first, first] = 1.
            for second in range(self.nb_strategies):
                if second != first:
                    fp = self.fixation_probability(second, first, beta, *args)
                    fixation_probabilities[first, second] = fp
                    tmp = fp / float(self.nb_strategies - 1)
                    transitions[first, second] = tmp
                    transitions[first, first] = transitions[first, first] - tmp

        return transitions.transpose(), fixation_probabilities

    def calculate_stationary_distribution(self, beta: float, *args: Optional[list]) -> np.ndarray:
        """
        Calculates the stationary distribution of the monomorphic states is mu = 0 (SML).
        Otherwise, it calculates the stationary distribution including all possible population states.

        This function is recommended only for Hermitian transition matrices.

        Parameters
        ----------
        beta : float
            intensity of selection.
        args : Optional[list]
            extra arguments for calculating payoffs.
        Returns
        -------
        numpy.ndarray
            A vector containing the stationary distribution
        """
        if self.mu > 0:
            t = self.calculate_full_transition_matrix(beta, *args).toarray()
        else:
            t, _ = self.transition_and_fixation_matrix(beta, *args)

        # Check if there is any transition with value 1 - this would mean that the game is degenerate
        if np.isclose(t, 1., atol=1e-11).any():
            warn(
                "Some of the entries in the transition matrix are close to 1 (with a tolerance of 1e-11). "
                "This could result in more than one eigenvalue of magnitude 1 "
                "(the Markov Chain is degenerate), so please be careful when analysing the results.", RuntimeWarning)

        # noinspection PyTupleAssignmentBalance
        eigenvalues, eigenvectors = eig(t)
        # calculate stationary distributions using eigenvalues and eigenvectors
        index_stationary = np.argmin(
            abs(eigenvalues - 1.0))  # look for the element closest to 1 in the list of eigenvalues
        sd = abs(eigenvectors[:, index_stationary].real)  # it is essential to access the matrix by column

        return sd / sd.sum()



"""
API reference documentation for `behaviors` submodule.
"""

import egttools.behaviors.CRD as CRD
import egttools.behaviors.NormalForm as NormalForm
import egttools.behaviors.pgg_behaviors as pgg_behaviors

__all__ = ['CRD', 'NormalForm', 'pgg_behaviors']



"""
API reference documentation for `behaviors.NormalForm.TwoActions` submodule.
"""

try:
    from egttools.numerical.numerical_.behaviors.NormalForm.TwoActions import (ActionInertia, Cooperator, Defector,
                                                                               GRIM,
                                                                               GenerousTFT,
                                                                               GradualTFT, ImperfectTFT, Pavlov, Random,
                                                                               SuspiciousTFT,
                                                                               TFT, TFTT, TTFT, )
except Exception:
    raise Exception("numerical package not initialized")
else:
    from .nfg_strategies import EpsilonTFT, EpsilonGRIM, Detective, MemoryOneStrategy

__all__ = ['ActionInertia', 'Cooperator', 'Defector', 'GRIM', 'GenerousTFT', 'GradualTFT', 'ImperfectTFT', 'Pavlov',
           'Random', 'SuspiciousTFT', 'TFT', 'TFTT', 'TTFT', 'EpsilonTFT', 'EpsilonGRIM', 'Detective',
           'MemoryOneStrategy']



"""Custom data structures used to store data from numerical simulations."""

try:
    from ..numerical.numerical_.DataStructures import DataTable
except Exception:
    raise Exception("numerical package not initialized")

__all__ = ['DataTable']



import pytest
import numpy as np, networkx as nx, egttools as egt, matplotlib.pyplot as plt


def test_plot_gradients_2d_array():
    gradients = np.random.rand(10, 2)
    # all ok returns None
    assert isinstance(egt.plotting.plot_gradients(gradients), plt.Axes)


def test_plot_gradient_none_input():
    # should error on input not being a np.array
    # accessing a particular atribute
    with pytest.raises(AttributeError):
        egt.plotting.plot_gradients(None)


def test_draw_invasion_diagram():
    # preliminary from the examples
    T, R, P, S, beta, Z = 2, 1, 0, 1, 1, 100
    A = np.array([[P, T], [S, R]])

    strategies = [
        egt.behaviors.NormalForm.TwoActions.Cooperator(),
        egt.behaviors.NormalForm.TwoActions.Random(),
        egt.behaviors.NormalForm.TwoActions.GRIM(),
    ]

    strategy_labels = [
        strategy.type().replace("NFGStrategies::", "") for strategy in strategies
    ]

    game = egt.games.NormalFormGame(1, A, strategies)
    evolver = egt.analytical.PairwiseComparison(Z, game)
    (
        transition_matrix,
        fixation_probabilities,
    ) = evolver.calculate_transition_and_fixation_matrix_sml(beta)

    stationary_distribution = egt.utils.calculate_stationary_distribution(
        transition_matrix.transpose()
    )

    assert isinstance(
        egt.plotting.draw_invasion_diagram(
            strategies=strategy_labels,
            drift=1,
            fixation_probabilities=fixation_probabilities,
            stationary_distribution=stationary_distribution,
        ),
        nx.DiGraph,
    )



from egttools.distributions import multivariate_hypergeometric_pdf
from egttools import sample_simplex, calculate_nb_states, USES_BOOST
from scipy.stats import multivariate_hypergeom
import numpy as np
import pytest


@pytest.mark.skipif(USES_BOOST is False, reason="requires egttools to be compiled with Boost C++")
def test_multivariate_hypergeometric_distribution_two_strategies():
    population_size = 500
    group_size = 10
    nb_strategies = 2
    nb_states = calculate_nb_states(population_size, nb_strategies)
    nb_group_combinations = calculate_nb_states(group_size, nb_strategies)

    for state_index in range(nb_states):
        state = sample_simplex(state_index, population_size, nb_strategies)
        for group_index in range(nb_group_combinations):
            group = sample_simplex(group_index, group_size, nb_strategies)
            prob_scipy = multivariate_hypergeom.pmf(group, state, group_size)
            prob_egttools = multivariate_hypergeometric_pdf(population_size,
                                                            nb_strategies,
                                                            group_size,
                                                            group,
                                                            state)

            assert np.isclose(prob_egttools, prob_scipy)


@pytest.mark.skipif(USES_BOOST is False, reason="requires egttools to be compiled with Boost C++")
def test_multivariate_hypergeometric_distribution_three_strategies():
    population_size = 50
    group_size = 4
    nb_strategies = 3
    nb_states = calculate_nb_states(population_size, nb_strategies)
    nb_group_combinations = calculate_nb_states(group_size, nb_strategies)

    for state_index in range(nb_states):
        state = sample_simplex(state_index, population_size, nb_strategies)
        for group_index in range(nb_group_combinations):
            group = sample_simplex(group_index, group_size, nb_strategies)
            prob_scipy = multivariate_hypergeom.pmf(group, state, group_size)
            prob_egttools = multivariate_hypergeometric_pdf(population_size,
                                                            nb_strategies,
                                                            group_size,
                                                            group,
                                                            state)

            assert np.isclose(prob_egttools, prob_scipy)



import pytest

import os
from sys import platform

import numpy as np

egt = pytest.importorskip("egttools")

Random = egt.Random
PairwiseComparisonNumerical = egt.numerical.PairwiseComparisonNumerical
NormalFormGame = egt.games.NormalFormGame


@pytest.fixture
def setup_hawk_dove_parameters() -> np.ndarray:
    # Payoff matrix
    v, d, t = 2, 3, 1
    payoffs = np.array([
        [(v - d) / 2, v],
        [0, (v / 2) - t],
    ])

    # Necessary to avoid issues with Anaconda on MacOSX
    if platform == "darwin":
        os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'

    Random.init(3610063510)

    assert Random._seed == 3610063510

    return payoffs


def test_normal_form_game_runs(setup_hawk_dove_parameters) -> None:
    """
    This test checks that a Normal form game runs as expected
    """

    payoffs = setup_hawk_dove_parameters

    game = NormalFormGame(1, payoffs)

    assert game.nb_strategies() == 2
    assert game.nb_rounds == 1
    assert game.type() == "NormalFormGame"
    np.testing.assert_array_equal(game.payoffs(), payoffs)
    np.testing.assert_array_equal(game.expected_payoffs(), payoffs)


def test_pairwise_moran_run(setup_hawk_dove_parameters) -> None:
    """
    This test checks that the run method of PairwiseComparisonNumerical executes.
    """
    payoffs = setup_hawk_dove_parameters

    game = NormalFormGame(1, payoffs)

    pop_size = 100
    cache_size = 10000
    nb_generations = int(1e6)
    beta = 1
    mu = 1e-3
    initial_state = [50, 50]

    evolver = PairwiseComparisonNumerical(pop_size, game, cache_size)
    result = evolver.run(nb_generations, beta, mu, initial_state)

    assert result.shape == (nb_generations + 1, game.nb_strategies())


def test_pairwise_moran_stationary_distribution(setup_hawk_dove_parameters) -> None:
    """
    This test checks that the stationary_distribution method of PairwiseComparisonNumerical executes.
    """
    payoffs = setup_hawk_dove_parameters

    game = NormalFormGame(1, payoffs)

    pop_size = 100
    cache_size = 10000
    nb_generations = int(1e3)
    transitory = int(1e3)
    beta = 10
    mu = 1e-3
    runs = 10

    nb_states = pop_size + 1

    evolver = PairwiseComparisonNumerical(pop_size, game, cache_size)
    dist = evolver.estimate_stationary_distribution(runs, nb_generations, transitory, beta, mu)
    assert dist.shape == (nb_states,)


def test_pairwise_moran_stationary_distribution_sparse(setup_hawk_dove_parameters) -> None:
    """
    This test checks that the stationary_distribution method of PairwiseComparisonNumerical executes.
    """
    payoffs = setup_hawk_dove_parameters

    game = NormalFormGame(1, payoffs)

    pop_size = 100
    cache_size = 10000
    nb_generations = int(1e3)
    transitory = int(1e3)
    beta = 10
    mu = 1e-3
    runs = 10

    nb_states = pop_size + 1

    evolver = PairwiseComparisonNumerical(pop_size, game, cache_size)
    dist = evolver.estimate_stationary_distribution_sparse(runs, nb_generations, transitory, beta, mu)
    assert dist.shape == (1, nb_states)



from typing import Union, List

import numpy as np

from . import AbstractNPlayerGame
from .. import sample_simplex, calculate_state
from ..behaviors.CPR.abstract_cpr_strategy import AbstractCPRStrategy

from scipy.stats import multivariate_hypergeom


class NPlayerStagHunt(AbstractNPlayerGame):
    """
    This game is based on the article
    Pacheco et al., ‘Evolutionary Dynamics of Collective Action in N -Person Stag Hunt Dilemmas’.
    """

    def __init__(self, group_size, enhancement_factor, cooperation_threshold, cost):
        super().__init__(2, group_size)  # Must be initialized at the beginning!
        self.group_size_ = group_size  # N
        self.enhancement_factor_ = enhancement_factor  # F
        self.cooperation_threshold_ = cooperation_threshold  # M
        self.cost_ = cost  # c
        self.strategies = ['Defect', 'Cooperate']

        self.nb_strategies_ = 2
        self.nb_group_configurations_ = self.nb_group_configurations()

        self.calculate_payoffs()

    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        """
        Calculate the payoff of each strategy inside the group.

        $\\Pi_{D}(k) = (kFc)\theta(k-M)$
        $\\Pi_{C}(k) = \\Pi_{D}(k) - c$

        Parameters
        ----------
        group_composition: Union[List[int], numpy.ndarray]
            counts of each strategy inside the group.
        game_payoffs: numpy.ndarray
            container for the payoffs of each strategy

        """
        if group_composition[0] == 0:
            game_payoffs[0] = 0
            game_payoffs[1] = self.cost_ * (self.enhancement_factor_ - 1)
        elif group_composition[1] == 0:
            game_payoffs[0] = 0
            game_payoffs[1] = 0
        else:
            game_payoffs[0] = ((group_composition[1]
                                * self.enhancement_factor_)
                               / self.group_size_) if group_composition[
                                                          1] >= self.cooperation_threshold_ else 0  # Defectors
            game_payoffs[1] = game_payoffs[0] - self.cost_  # Cooperators

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_group_configurations_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()

    def __str__(self) -> str:
        string = f'''
        Python implementation the N-player Stag Hunt Game.\n
        Game parameters
        -------
        N = {self.group_size_}\n
        F = {self.enhancement_factor_}\n
        M = {self.cooperation_threshold_}\n
        cost = {self.cost_}\n
        '''
        return string

    def type(self) -> str:
        return "NPlayerStagHunt"

    def save_payoffs(self, file_name: str) -> None:
        with open(file_name, 'w') as f:
            f.write('Payoffs for each type of player and each possible state:\n')
            f.write(f'rows: {" ,".join([strategy for strategy in self.strategies_])}\n')
            f.write('cols: all possible group compositions starting at (0, 0, ..., group_size)\n')
            f.write(f'{self.payoffs()}')
            f.write(f'N = {self.group_size_}\n')
            f.write(f'F = {self.enhancement_factor_}\n')
            f.write(f'M = {self.cooperation_threshold_}\n')
            f.write(f'cost = {self.c_}\n')


class CommonPoolResourceDilemma(AbstractNPlayerGame):
    def __init__(self, group_size: int, a: float, b: float, min_e: float, cost: float, nb_strategies: int = 3,
                 strategies: List[int] = None):
        AbstractNPlayerGame.__init__(self, nb_strategies, group_size)  # Must be initialized before anything else!

        self.group_size_ = group_size  # N
        self.a_ = a
        self.b_ = b
        self.min_e = min_e
        self.cost_ = cost
        self.strategies_ = strategies
        self.nb_strategies_ = nb_strategies

        self.nb_group_configurations_ = self.nb_group_configurations()

        self.calculate_payoffs()

    def calculate_total_extraction(self, group_composition):
        extraction = 0

        for i, nb_players_with_strategy_i in enumerate(group_composition):
            extraction += self.extraction_strategy(i) * nb_players_with_strategy_i
        return extraction

    def extraction(self, x, x_total):
        cost = 0
        if x <= self.min_e:
            cost = self.cost_
        return ((x / x_total) * ((self.a_ * x_total) - (self.b_ * x_total * x_total))) - cost

    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        # Each player's payoff is a function of the value of theirs and other extractiers' extraction
        x_total = self.calculate_total_extraction(group_composition)

        for i, strategy_count in enumerate(group_composition):
            if strategy_count == 0:
                game_payoffs[i] = 0
            else:
                x = self.extraction_strategy(i)
                game_payoffs[i] = self.extraction(x, x_total)  # UPDATED: ONLY MARKET 2

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_group_configurations_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()

    def extraction_strategy(self, strategy_id):
        if self.strategies_ is not None:
            return self.strategies_[strategy_id]
        else:
            extraction = 0
            if strategy_id == 0:
                extraction = self.min_extraction()
            elif strategy_id == 1:
                extraction = self.low_extraction()
            elif strategy_id == 2:
                extraction = self.high_extraction()
            return extraction

    def high_extraction(self):
        return self.a_ / (self.b_ * self.group_size_)

    def low_extraction(self):
        group_optimal = self.a_ / (2 * self.b_)  # CHECK
        return group_optimal / self.group_size_

    def min_extraction(self):
        return self.min_e

    def conditional_low(self, x_total):
        x = 0
        if x_total / self.resource_state_ > self.belief_low_:
            x = self.low_extraction()
        else:
            x = self.high_extraction()
        return x


class CommonPoolResourceDilemmaCommitment(AbstractNPlayerGame):
    def __init__(self, group_size: int, a: float, b: float, cost: float, fine: float,
                 strategies: List[AbstractCPRStrategy]):
        """
        Common Pool resource game with commitment, but the threshold is set as another strategy
        Parameters
        ----------
        group_size: int
            Size of the group playing the game.
        a and b: flaot
            Both a and b are the parameters of the game, if a>>b, then it's attractive to invest in the CPR
        cost:
            cost of making the commitment for commitment proposing strategy
        fine: float
            fine for defection when accepting the commitment
        strategies: Lit[int]
            List of strategies which will play the game
        """
        AbstractNPlayerGame.__init__(self, len(strategies), group_size)  # Must be initialized before anything else!
        self.group_size_ = group_size

        self.a_ = a
        self.b_ = b
        self.fine_ = fine
        self.cost_ = cost
        self.strategies_ = strategies
        self.nb_strategies_ = len(strategies)

        self.nb_group_configurations_ = self.nb_group_configurations()
        self.extractions_ = np.zeros(self.nb_group_configurations_, dtype=np.float64)

        self.calculate_payoffs()

    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        # Each player's payoff is a function of the value of theirs and other extractiers' extraction
        commitment_accepted = np.zeros(shape=(self.nb_strategies_,), dtype=bool)
        nb_committed = self.get_nb_committed(group_composition)
        self.check_if_commitment_validated(nb_committed, group_composition, commitment_accepted)
        group_extraction = self.calculate_total_extraction(commitment_accepted, group_composition)

        for index, strategy_count in enumerate(group_composition):
            if strategy_count == 0:
                game_payoffs[index] = 0
            else:
                extraction = self.strategies_[index].get_extraction(self.a_, self.b_, self.group_size_,
                                                                    commitment_accepted[index])
                game_payoffs[index] = self.strategies_[index].get_payoff(self.a_, self.b_, extraction, group_extraction,
                                                                         self.fine_, self.cost_,
                                                                         commitment_accepted[index])
        self.add_group_extraction(group_extraction, group_composition)

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_group_configurations_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()

    def calculate_total_extraction(self, commitment_accepted: np.ndarray,
                                   group_composition: Union[List[int], np.ndarray]) -> float:
        extraction = 0
        for index, strategy_count in enumerate(group_composition):
            if strategy_count > 0:
                extraction += strategy_count * self.strategies_[index].get_extraction(self.a_, self.b_,
                                                                                      self.group_size_,
                                                                                      commitment_accepted[index])
        return extraction

    def get_nb_committed(self, group_composition: Union[List[int], np.ndarray]) -> int:
        nb_committed = 0
        for index, strategy_count in enumerate(group_composition):
            nb_committed += strategy_count * self.strategies_[index].would_like_to_commit()
        return nb_committed

    def check_if_commitment_validated(self, nb_committed: int, group_composition: Union[List[int]],
                                      commitment_accepted: np.ndarray) -> None:
        validated = False
        # A commitment will only be validated if a commitment proposing strategy is present in the group
        for index, strategy_count in enumerate(group_composition):
            if strategy_count > 0:
                if self.strategies_[index].proposes_commitment():
                    commitment_accepted[index] = self.strategies_[index].is_commitment_validated(nb_committed)
                    validated = commitment_accepted[index]
        # If there is a commitment proposing strategy in the group for which the commitment would be validated
        # Then we check for every other strategy if they would be part of the commitment
        if validated:
            for index, strategy_count in enumerate(group_composition):
                if strategy_count > 0:
                    if not self.strategies_[index].proposes_commitment():
                        commitment_accepted[index] = self.strategies_[index].is_commitment_validated(validated)

    def add_group_extraction(self, group_extraction, group_composition):
        self.extractions_[
            calculate_state(group_size=self.group_size_, group_composition=group_composition)] = group_extraction

    def calculate_expected_consumption(self, population_state: np.ndarray) -> float:

        # multivariate PDF
        rv = multivariate_hypergeom(population_state, self.group_size_)

        expected_extraction = 0.0
        # Iterate over all possible group compositions
        for i in range(self.nb_group_configurations_):
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            # Estimate probability of the current group composition
            expected_extraction += self.extractions_[i] * rv.pmf(x=group_composition)

        return expected_extraction



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from . import AbstractCRDStrategy


class GoalBasedCRDStrategy(AbstractCRDStrategy):
    def __init__(self, a0: int, s1r0: int, s1r1: int, s1r2: int, s1r3: int, s1r4: int,
                 s2r0: int, s2r1: int, s2r2: int, s2r3: int, s2r4: int, switch: int, group_size: int):
        """
        A CRD strategy which adapts in function of a moving average of the contributions of the rest of the group.

        Parameters
        ----------
        a0: int
            how much to contribute in the initial round.
        aa: int
            how much to contribute when you have contributed above the average of the group in the previous round.
        am: int
            how much to contribute when you have contributed equal to the average of the group in the previous round.
        ab: int
            how much to contribute when you have contributed below the average of the group in the previous round.
        group_size: int
            Size of the group.
        """
        super().__init__()
        self.s1_ = (s1r0, s1r1, s1r2, s1r3, s1r4)
        self.s2_ = (s2r0, s2r1, s2r2, s2r3, s2r4)
        self.a0_ = a0
        self.switch_ = switch
        self.last_action_ = self.a0_
        self.nb_opponents_ = group_size - 1
        self.public_account_ = 0

    def get_action(self, time_step: int, group_contributions_prev: int):
        avg = round(group_contributions_prev / self.nb_opponents_)

        if time_step == 0:
            self.public_account_ += self.a0_
            return self.a0_
        else:
            self.public_account_ += group_contributions_prev

            if self.public_account_ < self.switch_:  # S1
                self.last_action_ = self.s1_[avg]
            else:  # S2
                self.last_action_ = self.s2_[avg]

            # Update public account with the current action and return it
            self.public_account_ += self.last_action_
            return self.last_action_

    def type(self):
        return "GoalBasedCRDStrategy({},{},{},{},{},{},{},{},{},{},{},{})".format(self.a0_, *self.s1_, self.switch_,
                                                                                  *self.s2_)

    def __str__(self):
        return self.type()



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

import pytest

egttools = pytest.importorskip("egttools")


def test_calculate_nb_states():
    assert egttools.numerical.numerical.calculate_nb_states == egttools.calculate_nb_states


def test_calculate_state():
    assert egttools.numerical.numerical.calculate_state == egttools.calculate_state


def test_sample_simple():
    assert egttools.numerical.numerical.sample_simplex == egttools.sample_simplex


def test_abstract_game():
    assert egttools.numerical.numerical.games.AbstractGame == egttools.games.AbstractGame


def test_normal_form_game():
    assert egttools.numerical.numerical.games.NormalFormGame == egttools.games.NormalFormGame


def test_abstract_nfg_strategy():
    assert egttools.numerical.numerical.behaviors.NormalForm.AbstractNFGStrategy == egttools.behaviors.NormalForm.AbstractNFGStrategy



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""
This python module contains some utility functions
to find saddle points and plot gradients in 2 player, 2 strategy games.
"""
from warnings import warn

import numpy as np
from scipy.linalg import schur, eigvals
from scipy.sparse import csr_matrix, csc_matrix
from typing import Optional, List, Generator, Union, Callable, Dict
from networkx import Graph
from egttools.games import AbstractGame


def find_saddle_type_and_gradient_direction(gradient: Union[List[float], np.ndarray],
                                            saddle_points_idx: Union[List[int], np.ndarray],
                                            offset: float = 0.01):
    """
    Finds whether a saddle point is stable or not. And defines the direction of the
    gradient among stable and unstable points.

    Parameters
    ----------
    gradient : Union[List[float], numpy.ndarray[numpy.float64[1,m]]]
        array containing the gradient of selection for all states of the population
    saddle_points_idx : Union[List[int], numpy.ndarray[numpy.int64[1,m]]]
        array containing the saddle points indices
    offset : float
        offset for the gradient_directions, so that arrows don't overlap with point

    Returns
    -------
    Tuple[List[bool], List[float]]
        Tuple containing an array that indicates the type of saddle points and another array indicating
        the direction of the gradient between unstable and stable points
    """
    saddle_type = []
    gradient_direction = []
    nb_points = len(gradient)
    real_offset = offset * (nb_points - 1)
    for i, point in enumerate(saddle_points_idx):
        if point < nb_points - 1:
            if point > 0:
                if gradient[point + 1] > 0:
                    saddle_type.append(False)
                    gradient_direction.append((point, saddle_points_idx[i + 1] - real_offset))
                elif gradient[point - 1] < 0:
                    saddle_type.append(False)
                    gradient_direction.append((point, saddle_points_idx[i - 1] + real_offset))
                else:
                    saddle_type.append(True)
            else:
                if gradient[point + 1] > 0:
                    saddle_type.append(False)
                    gradient_direction.append((point, saddle_points_idx[i + 1] - real_offset))
                else:
                    saddle_type.append(True)
        else:
            if gradient[point - 1] < 0:
                saddle_type.append(False)
                gradient_direction.append((point, saddle_points_idx[i - 1] + real_offset))
            else:
                saddle_type.append(True)
    saddle_type = np.asarray(saddle_type)
    gradient_direction = np.asarray(gradient_direction) / (nb_points - 1)
    return saddle_type, gradient_direction


def get_payoff_function(strategy_i: int,
                        strategy_j: int,
                        nb_strategies: int,
                        game: AbstractGame) -> Callable[[int, int, Optional[List]], float]:
    """
    Returns a function which gives the payoff of strategy i against strategy j.

    The returned function will return the payoff of strategy i
    given k individuals of strategy i and group_size - k j strategists.

    Parameters
    ----------
    strategy_i : int
        index of strategy i
    strategy_j : int
        index of strategy j
    nb_strategies : int
        Total number of strategies in the population.
    game: egttools.games.AbstractGame
        A game object which contains the method `egttools.games.AbstractGame.payoff` which returns the payoff of
        a strategy given a group composition.

    Returns
    -------
    Callable[[int, int, Optional[List]], float]
        A function which will return the payoff of strategy i,
        given k individuals of strategy i and group_size - k j strategists.
    """

    def get_payoff(k: int, group_size: int, *args: Optional[List]) -> float:
        """
        Returns the payoff given k individuals of strategy i and group_size - k j strategists.

        This function is useful when we can assume that there will only be two
        strategies in a group at any moment in time.

        Parameters
        ----------
        k : int
            Number of players adopting strategy i in the group.
        group_size: int
            Total size of the group.
        args: Optional[List]
            Extra arguments which may be required ot calculate the payoff

        Returns
        -------
        float
            The payoff of strategy i in a group with k i strategists and group_size - k
            j strategists.
        """
        if k > group_size:
            raise Exception("You have indicated a wrong group composition. k must be smaller or equal to group_size.")
        group_composition = np.zeros(shape=(nb_strategies,), dtype=int)
        if strategy_i != strategy_j:
            group_composition[strategy_i] = k
            group_composition[strategy_j] = group_size - k
        else:
            group_composition[strategy_i] = group_size
        return game.payoff(strategy_i, group_composition.tolist())

    return get_payoff


def transform_payoffs_to_pairwise(nb_strategies: int,
                                  game: AbstractGame) -> np.ndarray:
    """
    This function transform a payoff matrix in full format to a pairwise format.

    The transformation should only be done if it is possible to assume that there will always be
    at most 2 strategies in a group at a given time. An example of this would be when calculating
    the Small Mutation Limit (SML) of the Pairwise Moran Process. In this case, we do not need to
    know the payoffs for each strategy for any group composition (i.e., when there are more than
    2 strategies in the group), but only for all possible combinations of each 2 strategies.

    To be able to represent this in a nb_strategies x nb_strategies square matrix, we make each
    entry of the matrix a function, which will return the payoff of the strategy given k players
    adopting strategy i and N - k players adopting strategy j.

    Parameters
    ----------
    nb_strategies : int
        Number of strategies in the population
    game : egttools.games.AbstractGame
        A game object which implements the method `egttools.games.AbstractGame.payoff` which returns the payoff of
        a strategy given a group composition.

    Returns
    -------
    numpy.ndarray[numpy.float64[m,m]]
        Returns the payoff matrix in shape nb_strategies x nb_strategies, and each entry of the payoff
        matrix is a function which will return the payoff of a strategy i against strategy j given
        a group composition with k members of strategy i and N - k members of strategy j.
    """
    return np.asarray(
        [[get_payoff_function(i, j, nb_strategies, game) for j in range(nb_strategies)] for i in range(nb_strategies)])


def calculate_stationary_distribution(transition_matrix: Union[np.ndarray, csr_matrix, csc_matrix]) -> np.ndarray:
    """
    Calculates stationary distribution from a transition matrix of Markov chain.

    The use of this function is not recommended if the matrix is non-Hermitian. Please use
    calculate_stationary_distribution_non_hermitian instead in this case.

    The stationary distribution is the normalized eigenvector associated with the eigenvalue 1

    Parameters
    ----------
    transition_matrix : Union[numpy.ndarray, scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]
        A 2 dimensional transition matrix

    Returns
    -------
    numpy.ndarray
        A 1-dimensional vector containing the stationary distribution

    See Also
    --------
    egttools.utils.calculate_stationary_distribution_non_hermitian

    """
    if (type(transition_matrix) == csr_matrix) or (type(transition_matrix) == csc_matrix):
        tmp = transition_matrix.toarray()
    else:
        tmp = transition_matrix

    # Check if there is any transition with value 1 - this would mean that the game is degenerate
    if np.isclose(tmp, 1., atol=1e-11).any():
        warn(
            "Some of the entries in the transition matrix are close to 1 (with a tolerance of 1e-11). "
            "This could result in more than one eigenvalue of magnitute 1 "
            "(the Markov Chain is degenerate), so please be careful when analysing the results.", RuntimeWarning)

    # calculate stationary distributions using eigenvalues and eigenvectors
    eigenvalues, eigenvectors = np.linalg.eig(tmp)
    index_stationary = np.argmin(abs(eigenvalues - 1.0))  # look for the element closest to 1 in the list of eigenvalues
    sd = abs(eigenvectors[:, index_stationary].T.real)  # it is essential to access the matrix by column
    return sd / sd.sum()  # normalize


def calculate_stationary_distribution_non_hermitian(
        transition_matrix: Union[np.ndarray, csr_matrix, csc_matrix]) -> np.ndarray:
    """
    Calculates stationary distribution from a transition matrix of Markov chain which is not hermitian.

    The stationary distribution is the normalized eigenvector associated with the eigenvalue 1

    Parameters
    ----------
    transition_matrix : Union[numpy.ndarray, scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]
        A 2 dimensional transition matrix

    Returns
    -------
    numpy.ndarray
        A 1-dimensional vector containing the stationary distribution

    See Also
    --------
    egttools.utils.calculate_stationary_distribution

    """
    if (type(transition_matrix) == csr_matrix) or (type(transition_matrix) == csc_matrix):
        tmp = transition_matrix.toarray()
    else:
        tmp = transition_matrix

    # Check if there is any transition with value 1 - this would mean that the game is degenerate
    if np.isclose(tmp, 1., atol=1e-11).any():
        warn(
            "Some of the entries in the transition matrix are close to 1 (with a tolerance of 1e-11). "
            "This could result in more than one eigenvalue of magnitute 1 "
            "(the Markov Chain is degenerate), so please be careful when analysing the results.", RuntimeWarning)

    # calculate stationary distributions using eigenvalues and eigenvectors
    # noinspection PyTupleAssignmentBalance
    schur_form, eigenvectors = schur(tmp)
    eigenvalues = eigvals(schur_form)
    index_stationary = np.argmin(abs(eigenvalues - 1.0))  # look for the element closest to 1 in the list of eigenvalues
    sd = abs(eigenvectors[:, index_stationary].T.real)  # it is essential to access the matrix by column
    return sd / sd.sum()  # normalize


def combine(values: List[Union[int, str]], length: int) -> Generator:
    """
    Outputs a generator that will generate an ordered list
    with the possible combinations of values with length.

    Each time the generator is called it will output a list
    of length :param length which contains a combinations of the elements in
    the list of values.

    Parameters
    ----------
    values : List
        elements to combine
    length : int
        size of the output

    Returns
    -------
    Generator
        A generator which outputs ordered combinations of value as a list of size
        length


    Examples
    --------
    >>> for value in combine([1, 2], 2):
    ...     print(value)
    [1, 1]
    [2, 1]
    [1, 2]
    [2, 2]
    """
    output = [values[0]] * length
    max_combinations = len(values) ** length
    yield output
    for i in range(1, max_combinations):
        for j in range(length):
            output[j] = values[(i // len(values) ** j) % len(values)]
        yield output


def calculate_nb_unique_combinations(slots_per_bin: Union[List[int], np.ndarray]) -> int:
    """
    Calculates the number of unique combinations given the required number
    of elements of each group, which should be given in List format in the
    slots_per_bin parameter.

    Parameters
    ----------
    slots_per_bin: Union[List[int], numpy.ndarray]
        The list should contain the required number of elements of each group
        that should be combined in a tuple of length sum(slots_per_bin).

    Returns
    -------
    int
        The total number of unique combinations of the groups in the available slots.

    """
    total_slots = np.sum(slots_per_bin)

    divider = np.prod([np.math.factorial(x) for x in slots_per_bin])

    return np.math.factorial(total_slots) // divider


def transform_graph_to_dict_of_lists(graph: Graph) -> Dict[int, List[int]]:
    graph_dictionary = {}
    for node in graph.nodes:
        graph_dictionary[node] = list(graph[node].keys())

    return graph_dictionary



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""Helper function to visualize evolutionary dynamics"""

import networkx as nx
import numpy
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.ticker import AutoMinorLocator
from typing import List, Optional, Tuple, Union


def plot_gradients(gradients: numpy.ndarray, fig_title: Optional[str] = None,
                   xlabel: Optional[str] = 'frequency of strategy 0',
                   ylabel: Optional[str] = 'gradient of selection (G)',
                   figsize: Optional[Tuple[int, int]] = (7, 5),
                   color: Optional[Optional[Union[str, Tuple[int, int, int], List[str], List[
                       Tuple[int, int, int]], plt.cm.colors.Colormap]]] = 'b',
                   linelabel: Optional[str] = None,
                   linewidth_gradient: Optional[Union[int, List[int]]] = 3,
                   marker: Optional[str] = None,
                   marker_plot_freq: Optional[int] = 1,
                   marker_linewidth: Optional[int] = 3,
                   marker_size: Optional[int] = 20,
                   marker_facecolor: Optional[Optional[Union[str, Tuple[int, int, int], List[str], List[
                       Tuple[int, int, int]], plt.cm.colors.Colormap]]] = 'b',
                   marker_edgecolor: Optional[Optional[Union[str, Tuple[int, int, int], List[str], List[
                       Tuple[int, int, int]], plt.cm.colors.Colormap]]] = 'b',
                   roots: Optional[List[np.ndarray]] = None,
                   stability: Optional[List[int]] = None,
                   linewidth_edges: Optional[Union[int, List[int]]] = 3,
                   edgecolors: Optional[str] = 'black',
                   nodesize: Optional[int] = 100,
                   nodeborder_width: Optional[int] = 3,
                   arrowstyle: Optional[str] = '-|>',
                   nb_minor_ticks: Optional[int] = 2,
                   major_ticks_length: Optional[int] = 7,
                   minor_ticks_length: Optional[int] = 4,
                   ticks_width: Optional[int] = 2,
                   ticks_labels_size: Optional[int] = 15,
                   ticks_direction: Optional[str] = 'in',
                   ticks_labels_pad: Optional[int] = 10,
                   ticks_labels_fontweight: Optional[Union[str, int]] = 'bold',
                   axis_labels_fontweight: Optional[Union[str, int]] = 'bold',
                   axis_labels_fontsize: Optional[int] = 15,
                   ticks_left: Optional[bool] = True,
                   ticks_right: Optional[bool] = True,
                   ticks_top: Optional[bool] = True,
                   ticks_bottom: Optional[bool] = True,
                   spine_left_linewidth: Optional[int] = 2,
                   spine_right_linewidth: Optional[int] = 2,
                   spine_top_linewidth: Optional[int] = 2,
                   spine_bottom_linewidth: Optional[int] = 2,
                   ax: Optional[plt.axis] = None) -> plt.axis:
    """
    This function plots the gradient of selection for 1-simplexes (only two strategies).

    There is the possibility of plotting the stationary points (roots) of the system and their stability,
    but it is recommended that you only do this when analysing the replicator equation.

    Parameters
    ----------
    gradients: a numpy array with the gradients to plot.
    fig_title: a title for the figure.
    xlabel: the label of the x axis.
    ylabel: the label of the y axis.
    figsize: the dimensions of the figure.
    color: the color to use to plot the line.
    linelabel: label assigned to the plotted line.
    linewidth_gradient: width of the gradient curve.
    marker: use a marker to plot the points (by default no marker is shown).
    marker_plot_freq: how often to plot a marker (so that there aren't many overlapping).
    marker_linewidth: linewidth of the edge of the marker.
    marker_size: size of the marker.
    marker_facecolor: marker fill color.
    marker_edgecolor: marker edge color.
    roots: a list of numpy arrays containing the coordinates of the stationary points of the dynamical system.
    stability: a list of integers indicating the stability of the roots (-1 - unstable, 0 - saddle, 1 - stable).
    linewidth_edges: width of the arrows indicating the direction of the gradients.
    edgecolors: color of the arrows indicating the direction of selection.
    nodesize: size of the circles representing the roots.
    nodeborder_width: width of the border of the circles.
    arrowstyle: style of the arrows that represent the direction of selection.
    nb_minor_ticks: number of minor ticks to display.
    major_ticks_length: length of major ticks.
    minor_ticks_length: length of minor ticks.
    ticks_width: width of the ticks.
    ticks_labels_size: size of the tick labels.
    ticks_direction: direction of the ticks ("in" or "out").
    ticks_labels_pad: pad for the labels of the ticks.
    ticks_labels_fontweight: font weight of the tick labels (e.g., "bold').
    axis_labels_fontweight: font weight of the axis labels (e.g., "bold').
    axis_labels_fontsize: font size of the axis labels.
    ticks_left: indicates whether to display ticks on the left spine.
    ticks_right: indicates whether to display ticks on the right spine.
    ticks_top: indicates whether to display ticks on the top spine.
    ticks_bottom: indicates whether to display ticks on the bottom spine.
    spine_left_linewidth: line width of the left spine.
    spine_right_linewidth: line width of the right spine.
    spine_top_linewidth: line width of the top spine.
    spine_bottom_linewidth: line width of the bottom spine.
    ax: a matplotlib.pyplot axis object in which this figure will be plot. If None, then a new axis and figure will
        be created.

    Returns
    -------
    matplotlib.pyplot.axis

    The axis in which the figure has been plot.
    """
    x_values = np.linspace(0, 1, num=gradients.shape[0], dtype=np.float64)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if fig_title is not None:
        ax.set_title(fig_title)

    ax.plot(x_values, gradients, color=color, linewidth=linewidth_gradient, label=linelabel, zorder=1)
    if marker is not None:
        [ ax.scatter(x_values[::marker_plot_freq], gradient[::marker_plot_freq], 
marker=marker, s=marker_size, facecolors=marker_facecolor,
			   edgecolors=marker_edgecolor, linewidths=marker_linewidth, zorder=1.5) for gradient in gradients.T]
    # if stability is given, draw the stable points and arrows
    if stability is not None:
        if roots is None:
            raise Exception("To plot the stability you must also input the roots of the dynamical system.")

        # sort roots by first element, this will make the plotting of gradient arrows in
        # the egttools.plotting.indicators.plot_gradients work
        indexes = sorted(range(len(roots)), key=lambda k: roots[k][0])
        sorted_roots = [roots[i] for i in indexes]
        sorted_stability = [stability[i] for i in indexes]

        G = nx.DiGraph()
        G.add_nodes_from(np.arange(len(sorted_roots)))

        pos = {0: (sorted_roots[0][0], 0)}
        # Create edges
        for i in range(1, len(sorted_roots)):
            pos[i] = (sorted_roots[i][0], 0)

            if (sorted_stability[i] == -1) and (sorted_stability[i - 1] == 1):
                G.add_edge(i, i - 1)
            elif (sorted_stability[i] == -1) and (sorted_stability[i - 1] == 0):
                G.add_edge(i, i - 1)
            elif (sorted_stability[i] == -1) and (sorted_stability[i - 1] == -1):
                G.add_edge(i, i - 1)
                G.add_edge(i - 1, -1)
            elif (sorted_stability[i] == 0) and (sorted_stability[i - 1] == 1):
                G.add_edge(i, i - 1)
            elif (sorted_stability[i] == 0) and (sorted_stability[i - 1] == -1):
                G.add_edge(i - 1, i)
            elif (sorted_stability[i] == 1) and (sorted_stability[i - 1] == -1):
                G.add_edge(i - 1, i)
            elif (sorted_stability[i] == 1) and (sorted_stability[i - 1] == 0):
                G.add_edge(i - 1, i)

        def get_color(x):
            if x == -1:
                return 'white'
            elif x == 0:
                return 'gray'
            else:
                return 'black'

        node_colors = list(map(get_color, sorted_stability))

        nx.draw(G, pos=pos, edgecolors=edgecolors, linewidths=nodeborder_width, node_size=nodesize,
                node_color=node_colors, width=linewidth_edges,
                arrowstyle=arrowstyle,
                ax=ax)
        ax.set_xlim(-0.1, 1.1)
        ax.spines['left'].set_position(('axes', 0.083))
        ax.spines['right'].set_position(('axes', 0.917))
        ax.spines.top.set_bounds((0, 1))
        ax.spines.bottom.set_bounds((0, 1))
        ax.spines.left.set_zorder(0)
        ax.spines.right.set_zorder(0)
        ax.spines.top.set_zorder(0)
        ax.spines.bottom.set_zorder(0)
    else:
        ax.plot([0, 1], [0, 0], color='k', linewidth=linewidth_edges, zorder=1.5)
        ax.set_xlim(0, 1)

    ax.axis('on')  # turns on axis
    # ax.set_xticks(np.linspace(0, 1, 6))
    ax.xaxis.set_minor_locator(AutoMinorLocator(nb_minor_ticks))
    ax.yaxis.set_minor_locator(AutoMinorLocator(nb_minor_ticks))
    ax.tick_params(which='both', left=ticks_left, bottom=ticks_bottom, top=ticks_top, right=ticks_right, labelleft=True,
                   labelbottom=True,
                   width=ticks_width,
                   labelsize=ticks_labels_size, direction=ticks_direction, pad=ticks_labels_pad)
    ax.tick_params(which='major', length=major_ticks_length)
    ax.tick_params(which='minor', length=minor_ticks_length)
    ax.spines['left'].set_linewidth(spine_left_linewidth)
    ax.spines['right'].set_linewidth(spine_right_linewidth)
    ax.spines['top'].set_linewidth(spine_top_linewidth)
    ax.spines['bottom'].set_linewidth(spine_bottom_linewidth)

    for label in ax.get_xticklabels():
        label.set_fontweight(ticks_labels_fontweight)  # If change to 551, label will be bold-like
    for label in ax.get_yticklabels():
        label.set_fontweight(ticks_labels_fontweight)  # If change to 551, label will be bold-like

    ax.set_xlabel(xlabel, fontsize=axis_labels_fontsize, fontweight=axis_labels_fontweight)
    ax.set_ylabel(ylabel, fontsize=axis_labels_fontsize, fontweight=axis_labels_fontweight)

    return ax


def plot_gradient(x, gradients, saddle_points, saddle_type, gradient_direction, fig_title='', xlabel='', figsize=(5, 4),
                  **kwargs):
    """
    Creates a figure plotting the gradient of selection together with the saddle points,
    and the gradient arrows.

    :param x: vector containing the possible states in x axis. It must have the same length as gradient
    :param gradients: vector containing the gradient for each possible state
    :param saddle_points: vector containing all saddle points
    :param saddle_type: vector of booleans indicating whether the saddle point is stable
    :param gradient_direction: vector of points indicating the direction of the gradient
                               between unstable and stable saddle points
    :param fig_title: a string containing the title of the figure
    :param xlabel: label for x axis
    :param figsize: a tuple indicating the size of the figure
    :param kwargs: you may pass an axis object
    :returns a figure object
    """
    if 'ax' not in kwargs:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        ax = kwargs['ax']
    ax.plot(x, gradients, zorder=1)
    ax.plot([0, 1], [0, 0], 'k', zorder=1)
    # First we plot unstable points
    if saddle_points[~saddle_type].any():
        points1 = ax.scatter(x=saddle_points[~saddle_type],
                             y=np.zeros((len(saddle_points[~saddle_type], ))),
                             marker="o", color='k', s=80,
                             facecolors='white',
                             zorder=3)
        points1.set_clip_on(False)  # Plot points over axis
    # Then we plot stable points
    if saddle_points[saddle_type].any():
        points2 = ax.scatter(x=saddle_points[saddle_type],
                             y=np.zeros((len(saddle_points[saddle_type], ))),
                             marker="o",
                             color='k',
                             s=80,
                             zorder=3)
        points2.set_clip_on(False)
    # Plot arrows following the gradient
    for point in gradient_direction:
        ax.annotate("",
                    xy=(point[1], 0.0), xycoords='data',
                    xytext=(point[0], 0.0), textcoords='data',
                    arrowprops=dict(arrowstyle="->",
                                    connectionstyle="arc3"),
                    zorder=2
                    )
    # Set the axis style
    ax.set_xlim(0, 1)
    ax.set_xlabel(xlabel)
    ax.set_ylabel('$G(x)$')
    ax.set_title(fig_title)
    return ax


def draw_invasion_diagram(strategies: List[str], drift: float, fixation_probabilities: np.ndarray,
                          stationary_distribution: np.ndarray,
                          atol: float = 1e-4,
                          max_displayed_label_letters: Optional[int] = 4,
                          min_strategy_frequency: Optional[float] = -1,
                          node_size: Optional[int] = 4000,
                          font_size_node_labels: Optional[int] = 18,
                          font_size_edge_labels: Optional[int] = 14,
                          font_size_sd_labels: Optional[int] = 12,
                          display_node_labels: Optional[bool] = True,
                          display_edge_labels: Optional[bool] = True,
                          display_sd_labels: Optional[bool] = True,
                          node_labels_top_separation: Optional[float] = 0.15,
                          node_labels_bottom_separation: Optional[float] = - 0.2,
                          edge_width: Optional[int] = 2,
                          node_linewidth: Optional[float] = 0,
                          node_edgecolors: Optional[str] = None,
                          figsize: Optional[Tuple[int, int]] = (10, 10),
                          dpi: Optional[int] = 150,
                          colors: Optional[List[str]] = None,
                          ax: Optional[plt.axis] = None) -> nx.Graph:
    """
    Draws the markov chain for a given stationary distribution of monomorphic states.

    Parameters
    ----------
    strategies : List[str]
        Strategies and array of string labels for each strategy present in the population.
    drift : float
        drift = 1/pop_size
    fixation_probabilities : numpy.ndarray[float, 2]
        A matrix specifying the fixation probabilities.
    stationary_distribution : numpy.ndarray[float, 1]
        An array containing the stationary distribution (probability of each state in the system).
    atol : float
        The tolerance for considering a value equal to 1 (to detect wheter there is random drift). Default is 1e-4.
    max_displayed_label_letters : int
        Maximum number of letters of the strategy labels contained in the `strategies` List to
        be displayed.
    min_strategy_frequency: Optional[float]
        Minimum frequency of a strategy (its probability given by the stationary distribution)
        to be shown in the Graph.
    font_size_node_labels : Optional[int]
        Font size of the labels displayed inside each node.
    font_size_edge_labels : Optional[int]
        Font size of the labels displayed in each edge (which contain the fixation probabilities).
    font_size_sd_labels : Optional[int]
        Font size of the labels displayed beside each node containing the value of the stationary distribution.
    display_node_labels : Optional[bool]
        Indicates wether the node labels should be displayed.
    display_edge_labels : Optional[bool]
        Indicates wether the edge labels should be displayed.
    display_sd_labels : Optional[bool]
        Indicates whether the stationary distribution labels should be displayed.
    node_labels_top_separation : Optional[float]
        Gives the separation of node label for nodes with positive y (y > 0)
    node_labels_bottom_separation : Optional[float]
        Gives the separation of node label for nodes with negative y (y <= 0)
    edge_width : Optional[int]
        Width of the edge line.
    node_linewidth: Optional[float]
        Line width of node border
    node_edgecolors: Optional[str]
        Colors of node borders
    figsize : Optional[Tuple[int, int]]
        Size of the default figure (Only used if ax is not specified).
    dpi : Optional[int]
        Pixel density of the default plot
    node_size : Optional[int]
        Size of the nodes of the Graph to be plotted
    colors : Optional[List[str]]
        A list with the colors used to plot the nodes of the graph.
    ax : Optional[plt.axis]
        Axis on which to draw the graph.

    Returns
    -------
    networkx.Graph
        The graph depicting the Markov chain which represents the invasion dynamics.

    Notes
    -----
    If there are too many strategies, this function may not only take a lot of time to generate the Graph, but
    it will also not be easy to visualize. Also, you should only use this function when ploting the invasion
    diagram assuming the small mutation limit of the replication dynamics (SML).

    See Also
    --------
    plot_gradient

    Examples
    -------
    >>> import egttools as egt
    >>> import matplotlib.pyplot as plt
    >>> strategies = [egt.behaviors.NormalForm.TwoActions.Cooperator(), egt.behaviors.NormalForm.TwoActions.Defector(),
    ...               egt.behaviors.NormalForm.TwoActions.TFT(), egt.behaviors.NormalForm.TwoActions.Pavlov(),
    ...               egt.behaviors.NormalForm.TwoActions.Random(), egt.behaviors.NormalForm.TwoActions.GRIM()]
    >>> strategy_labels = [strategy.type().replace("NFGStrategies::", '') for strategy in strategies]
    >>> T=4; R=2; P=1; S=0; Z= 100; beta=1
    >>> A = np.array([
    ...     [P, T],
    ...     [S, R]
    ... ])
    >>> game = egt.games.NormalFormGame(100, A, strategies)
    >>> evolver = egt.analytical.StochDynamics(len(strategies), game.expected_payoffs(), Z)
    >>> sd = evolver.calculate_stationary_distribution(beta)
    >>> transitions, fixation_probabilities = evolver.transition_and_fixation_matrix(beta)
    >>> fig, ax = plt.subplots(figsize=(5, 5), dpi=150)
    >>> G = egt.plotting.draw_invasion_diagram(strategy_labels, 1/Z, fixation_probabilities, sd,
    ...     node_size=2000, min_strategy_frequency=0.00001, ax=ax)
    >>> plt.axis('off')
    >>> plt.show() # display
    """
    fixation_probabilities_normalized = fixation_probabilities / drift

    used_strategies = [strategy for i, strategy in enumerate(strategies) if
                       (stationary_distribution[i] > min_strategy_frequency)]

    used_strategies_idx = np.where(stationary_distribution > min_strategy_frequency)[0]

    q = len(used_strategies)

    G = nx.DiGraph()
    G.add_nodes_from(used_strategies)
    if colors is None:
        from seaborn import color_palette
        ncolors = color_palette("colorblind", q)
    else:
        ncolors = colors

    if ax is None:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

    for j in range(q):
        for i in range(q):
            if fixation_probabilities_normalized[used_strategies_idx[i], used_strategies_idx[j]] >= 1 - atol:
                G.add_edge(used_strategies[i], used_strategies[j],
                           weight=fixation_probabilities_normalized[used_strategies_idx[i], used_strategies_idx[j]])

    eselect = [(u, v) for (u, v, d) in G.edges(data=True) if d['weight'] > 1 + atol]
    eselect_labels = dict(((u, v), r"{0:.2f}$\rho_N$".format(d['weight']))
                          for (u, v, d) in G.edges(data=True) if d['weight'] > 1 + atol)
    edrift = [(u, v) for (u, v, d) in G.edges(data=True) if np.isclose(d['weight'], 1, atol=atol)]

    pos = nx.circular_layout(G)  # positions for all nodes

    # nodes
    nx.draw_networkx_nodes(G,
                           pos,
                           node_size=node_size,
                           node_color=ncolors,
                           linewidths=node_linewidth,
                           edgecolors=ncolors if node_edgecolors is None else node_edgecolors,
                           ax=ax)

    # edges
    nx.draw_networkx_edges(G,
                           pos,
                           node_size=node_size,
                           edgelist=eselect,
                           width=edge_width,
                           arrows=True,
                           arrowstyle='-|>', ax=ax)
    nx.draw_networkx_edges(G,
                           pos,
                           node_size=node_size,
                           edgelist=edrift,
                           width=edge_width,
                           alpha=0.5,
                           style='dashed',
                           arrows=False, ax=ax)

    # node labels
    if display_node_labels:
        nx.draw_networkx_labels(G, pos, {strat: strat[:max_displayed_label_letters] for strat in used_strategies},
                                font_size=font_size_node_labels, font_weight='bold', font_color='black', ax=ax)

    # edge labels
    if display_edge_labels:
        nx.draw_networkx_edge_labels(G, pos, edge_labels=eselect_labels, font_size=font_size_edge_labels, ax=ax)

    for i, (key, value) in enumerate(pos.items()):
        x, y = value
        if y > 0:
            value = node_labels_top_separation
        else:
            value = node_labels_bottom_separation
        if display_sd_labels:
            ax.text(x, y + value, s="{0:.2f}".format(stationary_distribution[used_strategies_idx[i]]),
                    horizontalalignment='center', fontsize=font_size_sd_labels)

    return G




"""
API reference documentation for the `plotting` submodule.
"""

from .indicators import plot_gradient, plot_gradients, draw_invasion_diagram
from .simplex2d import Simplex2D
from .simplified import plot_replicator_dynamics_in_simplex, plot_pairwise_comparison_rule_dynamics_in_simplex

__all__ = ['plot_gradient', 'plot_gradients', 'draw_invasion_diagram', 'Simplex2D',
           'plot_replicator_dynamics_in_simplex', 'plot_pairwise_comparison_rule_dynamics_in_simplex']



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from typing import List, Tuple


class Opinion:
    def __init__(self, opinion: float, tag: int):
        self.opinion = opinion
        self.tag = tag

    def get_action(self):
        pass

    @property
    def type(self):
        return "Opinion {}".format(self.opinion)


def opinion_factory(strategies: List[Tuple[float, int]]):
    return [Opinion(opinion, tag) for opinion, tag in strategies]



"""
API reference documentation for `behaviors.CRD` submodule.
"""
try:
    from egttools.numerical.numerical_.behaviors.CRD import AbstractCRDStrategy, CRDMemoryOnePlayer
except Exception:
    raise Exception("numerical package not initialized")
else:
    from .moving_average import MovingAverageCRDStrategy
    from .goal_based import GoalBasedCRDStrategy
    from .time_based import TimeBasedCRDStrategy

__all__ = ['AbstractCRDStrategy', 'CRDMemoryOnePlayer', 'MovingAverageCRDStrategy', 'GoalBasedCRDStrategy',
           'TimeBasedCRDStrategy']



import egttools as egt
from scipy.special import comb


def calculate_nb_states(pop_size, nb_strategies):
    return comb(pop_size + nb_strategies - 1, pop_size, exact=True)


def test_calculate_nb_states():
    assert egt.calculate_nb_states(10, 3) == 66
    assert egt.calculate_nb_states(10, 10) == 92378
    assert egt.calculate_nb_states(100, 11) == 46897636623981


def test_calculate_state():
    assert egt.calculate_state(2, [0, 1, 0, 1]) == 6
    assert egt.calculate_state(4, [0, 3, 0, 1]) == 22
    assert egt.calculate_state(4, [4, 0, 0, 0]) == 0
    assert egt.calculate_state(4, [0, 0, 0, 4]) == egt.calculate_nb_states(4, 4) - 1


def test_sample_simplex():
    pass


def test_sample_unit_simplex():
    pass


def test_calculate_strategies_distribution():
    pass


def test_random():
    pass



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""
The `behaviors.pgg_behaviors` submodule contains strategies which can
be used with `egttools.games.pgg` game.
"""

from typing import List


class PGGOneShotStrategy:
    def __init__(self, action: int) -> None:
        self.action = action

    def get_action(self) -> int:
        return self.action

    @property
    def type(self) -> str:
        return "PGGOneShotStrategy"


def player_factory(actions: List[int]) -> List[PGGOneShotStrategy]:
    return [PGGOneShotStrategy(action) for action in actions]



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from . import AbstractCRDStrategy


class MovingAverageCRDStrategy(AbstractCRDStrategy):
    def __init__(self, a0: int, aa: int, am: int, ab: int, group_size: int):
        """
        A CRD strategy which adapts in function of a moving average of the contributions of the rest of the group.

        Parameters
        ----------
        a0: int
            how much to contribute in the initial round.
        aa: int
            how much to contribute when you have contributed above the average of the group in the previous round.
        am: int
            how much to contribute when you have contributed equal to the average of the group in the previous round.
        ab: int
            how much to contribute when you have contributed below the average of the group in the previous round.
        group_size: int
            Size of the group.
        """
        super().__init__()
        self.a0_ = a0
        self.aa_ = aa
        self.am_ = am
        self.ab_ = ab
        self.last_action_ = self.a0_
        self.nb_opponents_ = group_size - 1

    def get_action(self, time_step: int, group_contributions_prev: int):
        avg = group_contributions_prev / self.nb_opponents_
        if time_step == 0:
            return self.a0_
        else:
            if avg < self.last_action_:
                self.last_action_ = self.ab_
                return self.ab_
            elif avg > self.last_action_:
                self.last_action_ = self.aa_
                return self.aa_
            else:
                self.last_action_ = self.am_
                return self.am_

    def type(self):
        return "MovingAverageCRDStrategy({},{},{},{})".format(self.a0_, self.aa_, self.am_, self.ab_)

    def __str__(self):
        return self.type()



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

from typing import List, Union
import numpy as np

from .. import (sample_simplex, )
from . import AbstractNPlayerGame


class OpinionGame(AbstractNPlayerGame):
    def __init__(self, group_size: int, peer_pressure_importance: float, peer_pressure_ratio: float,
                 opinion_values: List[float]) -> None:
        """
        Classical Public Goods game with only 2 possible contributions (o or cost).

        Parameters
        ----------
        group_size:
            Size of the group playing the game.
        peer_pressure_importance:
            Importance of being close in opinion to your group
        peer_pressure_ratio:
            Ratio at which the peer pressure is more important
        opinion_values:
            Value of each opinion
        """
        AbstractNPlayerGame.__init__(self, len(opinion_values), group_size)
        self.group_size_ = group_size
        self.peer_pressure_importance_ = peer_pressure_importance
        self.peer_pressure_ratio_ = peer_pressure_ratio
        self.opinion_values_ = opinion_values
        self.nb_strategies_ = len(opinion_values)
        self.strategies_ = np.arange(self.nb_strategies_)

        self.nb_group_configurations_ = self.nb_group_configurations()

        self.calculate_payoffs()

    def play(self, group_composition: Union[List[int], np.ndarray], game_payoffs: np.ndarray) -> None:
        # Each player's payoff is a function of the value of their opinion
        # and the distribution of opinions in the group (in function of the peer pressure importance)
        for i, strategy_count in enumerate(group_composition):
            if strategy_count == 0:
                game_payoffs[i] = 0
            else:
                group_composition[i] -= 1
                mean = np.average(self.strategies_, weights=group_composition)
                var = moment(self.strategies_, group_composition, mean, 2)
                group_composition[i] += 1
                # calculate distance of opinion to the group
                distance = np.abs(self.strategies_[i] - mean)
                game_payoffs[i] = (1 - self.peer_pressure_ratio_) * self.opinion_values_[
                    i] + self.peer_pressure_ratio_ * self.opinion_values_[i] * (
                                          1 - sigmoid(distance * (1 - var), self.peer_pressure_importance_))

    def calculate_payoffs(self) -> np.ndarray:
        payoffs_container = np.zeros(shape=(self.nb_strategies_,), dtype=np.float64)
        for i in range(self.nb_group_configurations_):
            # Get group composition
            group_composition = sample_simplex(i, self.group_size_, self.nb_strategies_)
            self.play(group_composition, payoffs_container)
            for strategy_index, strategy_payoff in enumerate(payoffs_container):
                self.update_payoff(strategy_index, i, strategy_payoff)
            # Reinitialize payoff vector
            payoffs_container[:] = 0

        return self.payoffs()

    def __str__(self) -> str:
        string = f'''
        Python implementation of a public goods game.\n
        Game parameters
        -------
        cost = {self.c_}\n
        multiplication_factor = {self.r_}\n
        Strategies
        -------
        nb_strategies = {self.nb_strategies_}\n
        strategies = {self.strategies_}\n
        '''
        return string

    def type(self) -> str:
        return "OpinionGame"

    def save_payoffs(self, file_name: str) -> None:
        with open(file_name, 'w') as f:
            f.write('Payoffs for each type of player and each possible state:\n')
            f.write(f'rows: {" ,".join([strategy.type() for strategy in self.strategies_])}\n')
            f.write('cols: all possible group compositions starting at (0, 0, ..., group_size)\n')
            f.write(f'{self.payoffs_}')
            f.write(f'group_size = {self.group_size_}\n')
            f.write(f'peer_pressure_importance = {self.peer_pressure_importance_}\n')
            f.write(f'opinion_values = {self.opinion_values_}\n')


def moment(x, counts, c, n):
    return np.sum(counts * (x - c) ** n) / np.sum(counts)


def sigmoid(x, temperature):
    return 1. / (1. + np.exp(-temperature * (x - 1)))



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>

"""
The code used in here has been adapted from https://github.com/YannickJadoul/Parselmouth/blob/master/setup.py
"""

import io
import os
import re
import shlex
import sys

try:
    from skbuild import setup
except ImportError:
    print("Please update pip to pip 10 or greater, or a manually install the PEP 518 requirements in pyproject.toml",
          file=sys.stderr)
    raise

from setuptools import find_packages


def find_version():
    with io.open(os.path.join(os.path.dirname(__file__), "cpp/src", "version.h"), encoding='utf8') as f:
        version_file = f.read()
    version_match = re.search(r"^#define EGTTOOLS_VERSION ([\da-z.]+)$", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    version=find_version(),
    # packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    packages=['egttools', 'egttools.numerical', 'egttools.numerical.structure', 'egttools.analytical',
              'egttools.plotting', 'egttools.games',
              'egttools.behaviors',
              'egttools.behaviors.CRD', 'egttools.behaviors.NormalForm', 'egttools.behaviors.NormalForm.TwoActions',
              'egttools.behaviors.CPR',
              'egttools.helpers',
              'egttools.distributions',
              'egttools.datastructures'
              ],
    package_dir={'egttools': "src/egttools",
                 'egttools.numerical': "src/egttools/numerical",
                 'egttools.numerical.structure': 'src/egttools/numerical/structure',
                 'egttools.analytical': "src/egttools/analytical",
                 'egttools.plotting': "src/egttools/plotting", 'egttools.games': "src/egttools/games",
                 'egttools.behaviors': "src/egttools/behaviors", 'egttools.behaviors.CRD': "src/egttools/behaviors/CRD",
                 'egttools.behaviors.NormalForm': "src/egttools/behaviors/NormalForm",
                 'egttools.behaviors.NormalForm.TwoActions': "src/egttools/behaviors/NormalForm/TwoActions",
                 'egttools.behaviors.CPR': "src/egttools/behaviors/CPR",
                 'egttools.helpers': "src/egttools/helpers",
                 'egttools.distributions': "src/egttools/distributions",
                 'egttools.datastructures': "src/egttools/datastructures"
                 },
    cmake_args=shlex.split(os.environ.get('EGTTOOLS_EXTRA_CMAKE_ARGS', '')),
    cmake_install_dir="src/egttools/numerical",
    cmake_with_sdist=True,
)



"""The `numerical` module contains functions and classes to simulate evolutionary dynamics in finite populations."""

try:
    import egttools.numerical.numerical_ as numerical
except Exception:
    raise Exception("numerical package not initialized")
else:
    from .numerical_ import PairwiseComparisonNumerical
    from .numerical_ import GeneralPopulationEvolver
    from .numerical_ import NetworkEvolver

__all__ = ['numerical', 'PairwiseComparisonNumerical', 'GeneralPopulationEvolver', 'NetworkEvolver']



# Copyright (c) 2019-2021  Elias Fernandez
#
# This file is part of EGTtools.
#
# EGTtools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EGTtools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EGTtools.  If not, see <http://www.gnu.org/licenses/>
import numpy
import numpy as np
from scipy.optimize import root
from scipy.linalg import eigvals
from typing import Tuple, List, Optional, Callable
from .sed_analytical import StochDynamics, replicator_equation, replicator_equation_n_player
from .. import sample_unit_simplex


def get_pairwise_gradient_from_replicator(i: int, j: int, x: float, nb_strategies: int, payoffs: np.ndarray,
                                          freq_array: Optional[np.ndarray] = None) -> float:
    """
    Calculate the gradient for strategy/type `i` at the edges of the simplex (when there are
    only two strategies in the population `i` and `j`).

    Parameters
    ----------
    i: int
        index of the strategy whose gradient we wish to calculate
    j: int
        index of the other strategy present in the population
    x: float
        frequency of i type
    nb_strategies: int
        total number of strategies in the population
    payoffs: numpy.ndarray
        payoff matrix that defines the expected  payoff of any give strategy against each other
    freq_array: Optional[numpy.ndarray]
        optional vector to store the frequencies of each strategy in the population

    Returns
    -------
    float
        The gradient of strategy i.

    """
    if freq_array is None:
        freq_array = np.zeros(shape=(nb_strategies,))
    else:
        freq_array[:] = 0

    freq_array[i] = x
    freq_array[j] = 1. - x

    return replicator_equation(freq_array, payoffs)[i]


def get_pairwise_gradient_from_replicator_n_player(i: int, j: int, x: float, nb_strategies: int, group_size: int,
                                                   payoffs: np.ndarray,
                                                   freq_array: Optional[np.ndarray] = None) -> float:
    """
    Calculate the gradient for strategy/type `i` at the edges of the simplex (when there are
    only two strategies in the population `i` and `j`).

    Parameters
    ----------
    i: int
        index of the strategy whose gradient we wish to calculate
    j: int
        index of the other strategy present in the population
    x: float
        frequency of i type
    nb_strategies: int
        total number of strategies in the population
    group_size: int
        size of the group
    payoffs: numpy.ndarray
        payoff matrix that defines the expected  payoff of any give strategy against each other
    freq_array: Optional[numpy.ndarray]
        optional vector to store the frequencies of each strategy in the population

    Returns
    -------
    float
        The gradient of strategy i.

    """
    if freq_array is None:
        freq_array = np.zeros(shape=(nb_strategies,))
    else:
        freq_array[:] = 0

    freq_array[i] = x
    freq_array[j] = 1. - x

    return replicator_equation_n_player(freq_array, payoffs, group_size)[i]


def check_if_there_is_random_drift(payoff_matrix: np.ndarray,
                                   population_size: Optional[int] = None,
                                   group_size: int = 2,
                                   beta: Optional[float] = None,
                                   nb_points: int = 10,
                                   atol: float = 1e-7
                                   ) -> List[Tuple[int, int]]:
    """
    Checks if there is random drift along the edge between two strategies in the simplex.

    Parameters
    ----------
    payoff_matrix: numpy.ndarray
        The square matrix of payoffs. If the game is pairwise (group_size = 2) then each entry
        represents the payoff of the row strategy vs the column strategy. If the group_size > 2, then
        each entry should be a function that will return the payoff of the row strategy in a group of size N
        with N-k members of the column strategy. If you only have a matrix where the columns
        represent all possible game states, then you can use the function `egttools.utils.transform_payoffs_to_pairwise`
        to get a matrix in the correct form.

    population_size: Optional[int]
        The size of the population. If this value is not given, we assume that
        we calculate the dynamics in infinite populations using the replicator_equation.

    group_size: int
        The size of the group. If you specify population size, you should also specify this value. By default we assume
        that the game is pairwise.

    beta: Optional[float]
        The intensity of selection.If you specify population size, you should also specify this value.

    nb_points: int
        Number of points for which to check the gradient. It is 10 by default.

    atol: float
        Tolerance to consider a value zero

    Returns
    -------
    List[Tuple[int, int]]
        A list of tuples indicating the undirected edged where there should be random drift.

    """
    # To check if there is random drift, the transition probabilities should be zero

    points = np.linspace(0, 1, nb_points)
    f = None

    if population_size is None:
        freq_array = np.zeros(shape=(payoff_matrix.shape[0]))

        if group_size == 2:
            def gradient_function(i, j, x):
                return get_pairwise_gradient_from_replicator(i, j, x, payoff_matrix.shape[0],
                                                             payoff_matrix,
                                                             freq_array)
        else:
            def gradient_function(i, j, x):
                return get_pairwise_gradient_from_replicator_n_player(i, j, x, payoff_matrix.shape[0],
                                                                      group_size,
                                                                      payoff_matrix,
                                                                      freq_array)

        f = gradient_function
    else:
        if beta is None:
            raise Exception("the beta parameter must be specified!")
        evolver = StochDynamics(payoff_matrix.shape[0], payoff_matrix, population_size, group_size)

        def gradient_function(i, j, x):
            return evolver.gradient_selection(np.floor(x * population_size).astype(np.int64),
                                              i, j,
                                              beta)

        f = gradient_function

    solutions = []
    for row_strategy in range(payoff_matrix.shape[0]):
        # we don't want to look at the case where j==i
        for col_strategy in range(row_strategy + 1, payoff_matrix.shape[0]):
            gradients = []
            for point in points:
                res = f(row_strategy, col_strategy, point)
                gradients.append(res)

            if np.allclose(gradients, 0., atol=atol):
                solutions.append((row_strategy, col_strategy))

    return solutions


def find_roots_and_stability(gradient_function: Callable[[np.ndarray], np.ndarray], nb_strategies: int,
                             nb_initial_random_points: int = 3,
                             atol: float = 1e-7,
                             atol_neg: float = 1e-4, atol_pos: float = 1e-4,
                             atol_zero: float = 1e-4,
                             tol_close_points: float = 1e-4,
                             method: str = 'hybr') -> Tuple[List[np.array], List[int]]:
    """
    Searches for the roots of the differential equation `gradient_function` and calculates the stability based
    on an estimate of the Jacobian. This estimate is often imprecise which leads to wrong results.

    Parameters
    ----------
    gradient_function: Callable[[np.ndarray], np.ndarray]
        function that returns a numpy.ndarray with the gradient of every strategy/type given a
        current population state.
    nb_strategies: int
        number of strategies/types present in the population.
    nb_initial_random_points: int
        number of random points to use as initial states for the root function. These are
        additional to the vertex of the simplex.
    atol: float
        tolerance for considering that a point is in the simplex.
    atol_neg: float
        tolerance to consider a value negative.
    atol_pos: float
        tolerance to consider a value positive.
    atol_zero: float
        tolerance to determine if a value is zero.
    tol_close_points: float
        tolerance for considering that two points are equal.
    method: str
        one of the options described in `scipy.optimize.root`
        (see https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html)

    Returns
    -------
    Tuple[List[np.array], List[int]]
        A tuple containing the list of roots and a list with 1 indicating stable points, 0 saddle points
        and -1 unstable points.

    """
    # we test all the vertex of the simplex and some random initial points
    initial_states = [[0 if i != j else 1 for i in range(nb_strategies)] for j in range(nb_strategies)]
    for i in range(nb_initial_random_points):
        initial_states.append(sample_unit_simplex(nb_strategies))

    initial_states = np.asarray(initial_states)

    roots = []
    stability = []

    for initial_state in initial_states:
        sol = root(gradient_function, initial_state, method=method, jac=False)

        if sol.success:
            v = sol.x
            if check_if_point_in_unit_simplex(v, atol):
                # only add new fixed points to list
                if not np.array([np.allclose(v, x, atol=tol_close_points) for x in roots]).any():
                    roots.append(v)

                    # now we check the stability of the roots using the jacobian
                    eigenvalues = eigvals(sol.fjac)
                    print(eigenvalues)
                    # If all eigenvalues are negatives or zero it's stable
                    if (eigenvalues.real < -atol_neg).all() or np.array(
                            [np.isclose(el, 0., atol=atol_zero) for el in
                             eigenvalues.real[eigenvalues.real > -atol_neg]]).all():
                        stability.append(1)
                    # If all eigenvalues are positive or zero it's unstable
                    elif (eigenvalues.real > atol_pos).all() or np.array(
                            [np.isclose(el, 0., atol=atol_zero) for el in
                             eigenvalues.real[eigenvalues.real < atol_pos]]).all():
                        stability.append(-1)
                    else:  # saddle point
                        # This is probably wrong, but let's first assume that if we reach here, the point is a saddle
                        stability.append(0)
                        # # we need to check the hessian matrix to find out if the point is a saddle
                        # eigenvalues, _ = np.linalg.eig(sol.hess)
                        # if (eigenvalues > 0).any() and (eigenvalues < 0).any():
                        #     stability.append(0)

    return roots, stability


def check_if_point_in_unit_simplex(point: np.ndarray, delta: float = 1e-12) -> bool:
    """
    Checks if a point (in barycentric coordinates) is inside the unit simplex.

    Parameters
    ----------
    point: numpy.ndarray
        The barycentric coordinates of the point.
    delta: float
        Tolerance to consider a point outside the unit simplex.

    Returns
    -------
    bool
        Whether the point is inside the unit simplex.

    """
    if not np.isclose(np.sum(point), 1., atol=1.e-2):
        return False

    if not np.all((point > -delta) & (point < 1 + delta)):  # only if fp in simplex
        return False

    return True


def calculate_gradients(population_states: np.ndarray,
                        gradient_function: Callable[[np.ndarray], np.ndarray]) -> np.ndarray:
    """
    Calculates the gradients of selection of each of the states given in `population_states`.

    Parameters
    ----------
    population_states: numpy.ndarray
        A numpy array of shape (m,n) where n is the number of strategies in the population and
        m the number of states for which the gradient should be calculated.
    gradient_function: Callable[[np.ndarray], np.ndarray]
        A function which accepts a vector of shape (n,) containing the frequencies of each
        strategy/type in the population, and returns another vector of shape (n,) containing
        the gradient for each strategy.

    Returns
    -------
    numpy.ndarray
        A numpy array of shape (m,n) containing the gradients for
        each of the input states given in `population_states`.

    """
    return np.array([gradient_function(population_states[i]) for i in range(population_states.shape[0])])


def find_roots(gradient_function: Callable[[np.ndarray], np.ndarray],
               nb_strategies: int, nb_initial_random_points: int = 3,
               atol: float = 1e-7, tol_close_points: float = 1e-4,
               method: str = 'hybr') -> List[np.ndarray]:
    """
    Searches for the roots of the given differential equation.

    Parameters
    ----------
    gradient_function: Callable[[np.ndarray], np.ndarray]
        function that returns a numpy.ndarray with the gradient of every strategy/type given a
        current population state.
    nb_strategies: int
        number of strategies/types present in the population.
    nb_initial_random_points: int
        number of random points to use as initial states for the root function. These are
        additional to the vertex of the simplex.
    atol: float
        tolerance for considering that a point is in the simplex.
    tol_close_points: float
        tolerance for considering that two points are equal.
    method: str
        one of the options described in `scipy.optimize.root`
        (see https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html)

    Returns
    -------
    List[numpy.ndarray]
        A list of tuples with the identified roots/stationary points.

    """
    # we test all the vertex of the simplex and some random initial points
    initial_states = [[0 if i != j else 1 for i in range(nb_strategies)] for j in range(nb_strategies)]
    for i in range(nb_initial_random_points):
        initial_states.append(sample_unit_simplex(nb_strategies))

    initial_states = np.asarray(initial_states)

    roots = []

    for initial_state in initial_states:
        sol = root(gradient_function, initial_state, method=method, jac=False)

        if sol.success:
            v = sol.x
            if check_if_point_in_unit_simplex(v, atol):
                # only add new fixed points to list
                if not np.array([np.allclose(v, x, atol=tol_close_points) for x in roots]).any():
                    roots.append(v)

    return roots


def check_replicator_stability_pairwise_games(stationary_points: List[numpy.ndarray], payoff_matrix: numpy.ndarray,
                                              atol_neg: float = 1e-4, atol_pos: float = 1e-4,
                                              atol_zero: float = 1e-4) -> List[int]:
    """
    Calculates the stability of the roots assuming that they are from a system governed by the replicator
    equation (this function uses the Jacobian of the replicator equation in pairwise games to calculate the
    stability).

    Parameters
    ----------
    stationary_points: List[numpy.ndarray]
        a list of stationary points (represented as numpy.ndarray).
    payoff_matrix: numpy.ndarray
        a payoff matrix represented as a numpy.ndarray.
    atol_neg: float
        tolerance to consider a value negative.
    atol_pos: float
        tolerance to consider a value positive.
    atol_zero: float
        tolerance to determine if a value is zero.

    Returns
    -------
    List[int]
        A list of integers indicating the stability of the stationary points for the replicator equation:
        1 - stable
        -1 - unstable
        0 - saddle

    """

    def fitness(i: int, x: np.ndarray):
        return np.dot(payoff_matrix, x)[i]

    # First we build a Jacobian matrix
    def jacobian(x: numpy.ndarray):
        ax = np.dot(payoff_matrix, x)
        avg_fitness = np.dot(x, ax)
        jac = [[x[i] * (payoff_matrix[i, j] - np.dot(x, payoff_matrix[:, j])) if i != j else (
                fitness(i, x) - avg_fitness + x[i] * (payoff_matrix[i, i] - np.dot(x, payoff_matrix[:, i]))) for i in
                range(len(x))] for j in range(len(x))]
        return np.asarray(jac)

    stability = []

    for point in stationary_points:
        # now we check the stability of the roots using the jacobian
        eigenvalues = eigvals(jacobian(point))
        # If all eigenvalues are negatives or zero it's stable
        if (eigenvalues.real < -atol_neg).all() or np.array(
                [np.isclose(el, 0., atol=atol_zero) for el in eigenvalues.real[eigenvalues.real > -atol_neg]]).all():
            stability.append(1)
        # If all eigenvalues are positive or zero it's unstable
        elif (eigenvalues.real > atol_pos).all() or np.array(
                [np.isclose(el, 0., atol=atol_zero) for el in eigenvalues.real[eigenvalues.real < atol_pos]]).all():
            stability.append(-1)
        else:  # saddle point
            # This is probably wrong, but let's first assume that if we reach here, the point is a saddle
            stability.append(0)
            # # we need to check the hessian matrix to find out if the point is a saddle
            # eigenvalues, _ = np.linalg.eig(sol.hess)
            # if (eigenvalues > 0).any() and (eigenvalues < 0).any():
            #     stability.append(0)

    return stability



from typing import Dict, List
from egttools.numerical.structure import Network, NetworkGroup, NetworkSync, NetworkGroupSync
from egttools.games import AbstractSpatialGame


def network_factory(nb_strategies: int, beta: float, mu: float, game: AbstractSpatialGame, cache_size: int,
                    node_list: List[Dict[int, List[int]]]) -> List[Network]:
    """
    Generates a list of Network objects from the list of node, neighbours dictionaries.

    Parameters
    ----------
    nb_strategies : int
        Number of strategies in the population
    beta : float
        Intensity of selection
    mu : float
        Mutation rate
    game : egttools.games.AbstractSpatialGame
        A game to associate with each network
    cache_size : int
        The size of the cache memory to use
    node_list : List[Dict[int, List[int]]
        A list of dictionaries containing the nodes and their neighbours

    Returns
    -------
    List[Network]
        A list of Network objects

    """
    network_list = []
    for i, node_dictionary in enumerate(node_list):
        network_list.append(Network(nb_strategies, beta, mu, node_dictionary, game, cache_size))

    return network_list


def network_group_factory(nb_strategies: int, beta: float, mu: float, game: AbstractSpatialGame, cache_size: int,
                          node_list: List[Dict[int, List[int]]]) -> List[Network]:
    """
    Generates a list of NetworkGroup objects from the list of node, neighbours dictionaries.

    Parameters
    ----------
    nb_strategies : int
        Number of strategies in the population
    beta : float
        Intensity of selection
    mu : float
        Mutation rate
    game : egttools.games.AbstractSpatialGame
        A game to associate with each network
    cache_size : int
        The size of the cache memory to use
    node_list : List[Dict[int, List[int]]
        A list of dictionaries containing the nodes and their neighbours

    Returns
    -------
    List[Network]
        A list of Network objects

    """
    network_list = []
    for i, node_dictionary in enumerate(node_list):
        network_list.append(NetworkGroup(nb_strategies, beta, mu, node_dictionary, game, cache_size))

    return network_list


def network_sync_factory(nb_strategies: int, beta: float, mu: float, game: AbstractSpatialGame, cache_size: int,
                         node_list: List[Dict[int, List[int]]]) -> List[Network]:
    """
    Generates a list of Network objects from the list of node, neighbours dictionaries.

    Parameters
    ----------
    nb_strategies : int
        Number of strategies in the population
    beta : float
        Intensity of selection
    mu : float
        Mutation rate
    game : egttools.games.AbstractSpatialGame
        A game to associate with each network
    cache_size : int
        The size of the cache memory to use
    node_list : List[Dict[int, List[int]]
        A list of dictionaries containing the nodes and their neighbours

    Returns
    -------
    List[Network]
        A list of Network objects

    """
    network_list = []
    for i, node_dictionary in enumerate(node_list):
        network_list.append(NetworkSync(nb_strategies, beta, mu, node_dictionary, game, cache_size))

    return network_list


def network_group_sync_factory(nb_strategies: int, beta: float, mu: float, game: AbstractSpatialGame,
                               cache_size: int,
                               node_list: List[Dict[int, List[int]]]) -> List[Network]:
    """
    Generates a list of NetworkGroup objects from the list of node, neighbours dictionaries.

    Parameters
    ----------
    nb_strategies : int
        Number of strategies in the population
    beta : float
        Intensity of selection
    mu : float
        Mutation rate
    game : egttools.games.AbstractSpatialGame
        A game to associate with each network
    cache_size : int
        The size of the cache memory to use
    node_list : List[Dict[int, List[int]]
        A list of dictionaries containing the nodes and their neighbours

    Returns
    -------
    List[Network]
        A list of Network objects

    """
    network_list = []
    for i, node_dictionary in enumerate(node_list):
        network_list.append(NetworkGroupSync(nb_strategies, beta, mu, node_dictionary, game, cache_size))

    return network_list
